import exc.object.*;
import exc.block.*;

/**
 * Algorithm to emit instructions for Block.
 */
public class EmitInst {

  EmitInstEnv env;	// link to EmitInstEnv

  // nested label
  final static int MAX_NEST = 100;
  int break_level;
  int break_labels[];
  int continue_level;
  int continue_labels[];
  int ret_label;

  EmitLIR emitLIR;

  /** constructor */
  public EmitInst(){
    break_labels = new int[MAX_NEST];
    continue_labels = new int[MAX_NEST];
    emitLIR = new EmitLIR(env);
  }

  /**
   * Generate code for Function
   * it take Function Block to convert LIR level block
   */
  void run(FuncDefBlock def, EmitInstEnv env){
    this.env = env;
    ret_label = env.genLabel();
    
    FunctionBlock fblock = def.getBlock();

    // assign offset to parameters
    emit(fblock.getBody());
    // generate return label
    
  }

  /** emit LIR codes for Statement */
  public void emit(Statement s){
    if(s == null) return;
    emitLIR.emitStatementExpr(s.getExpr());
  }

  /** emit LIR code for basic block */
  public void emit(BasicBlock bb){
    if(bb == null) return;
    for(Statement s = bb.getHead(); s != null; s = s.getNext())
      emit(s);
  }

  /** emit LIR code for the list of basic block */
  public void emit(BlockList b_list){
    if(b_list == null) return;
    // process id_list.

    
    /* process blocks in BlockList */
    for(Block b = b_list.getHead(); b != null; b = b.getNext()){
      emit(b);
    }
  }

  void emitGoto(int label){
    env.emit(Inst.Goto(label));
  }

  void emitLabel(int label){
    env.emit(Inst.Label(label));
  }

  /** emit LIRs code for Block. */
  public void emit(Block b){
    int l1,l2;

    switch(b.Opcode()){
    case Xcode.LIST:	/* simple block */
      emit(b.getBasicBlock());
      break;

    case Xcode.COMPOUND_STATEMENT:
      emit(b.getBody());
      break;

    case Xcode.IF_STATEMENT:
      if(b.getThenBody() != null){
	l1 = env.genLabel();
	emitBranch(b.getCondBBlock(),l1,false);
	emit(b.getThenBody());
	if(b.getElseBody() != null){
	  l2 = env.genLabel();
	  emitGoto(l2);
	  emitLabel(l1);
	  emit(b.getElseBody());
	  emitLabel(l2);
	} else {
	  emitLabel(l1);
	}
      } else {
	l1 = env.genLabel();
	emitBranch(b.getCondBBlock(),l1,true);
	emit(b.getElseBody());
	emitLabel(l1);
      }
      break;
	
    case Xcode.FOR_STATEMENT: 
      {
	int lab = env.genLabel();
	int exit_lab = env.genLabel();
	int cont_lab = env.genLabel();

	emit(b.getInitBBlock()); // init part 

	emitBranch(b.getCondBBlock(),exit_lab,false);
	emitLabel(lab);
	
	/* set current evniroment */
	break_labels[break_level++] = exit_lab;
	continue_labels[continue_level++] = cont_lab;
	emit(b.getBody());
	emitLabel(cont_lab);
	emit(b.getIterBBlock());
	emitBranch(b.getCondBBlock(),lab,true);
	
	emitLabel(exit_lab);
	break_level--;
	continue_level--;     /* OUT */
	break;
      }

    case Xcode.WHILE_STATEMENT: 
      {
	int lab = env.genLabel();
	int exit_lab = env.genLabel();
	int cont_lab = env.genLabel();

	emitBranch(b.getCondBBlock(),exit_lab,false);
	emitLabel(lab);

	/* set current evniroment */
	break_labels[break_level++] = exit_lab;
	continue_labels[continue_level++] = cont_lab;
	emit(b.getBody());
	emitLabel(cont_lab);
	emitBranch(b.getCondBBlock(),exit_lab,false);
	emitLabel(exit_lab);
	break_level--;
	continue_level--;     /* OUT */
	break;
      }

    case Xcode.DO_STATEMENT: 
      {
	int exit_lab = env.genLabel();
	int cont_lab = env.genLabel();
	int lab = env.genLabel();

	emitLabel(lab);

	/* set current evniroment */
	break_labels[break_level++] = exit_lab;
	continue_labels[continue_level++] = cont_lab;
	emit(b.getBody());
	emitLabel(cont_lab);
	emitBranch(b.getCondBBlock(),lab,true);
	emitLabel(exit_lab);
	break_level--;
	continue_level--;     /* OUT */
	break;
      }

    case Xcode.BREAK_STATEMENT:
      if(break_level <= 0) fatal("bad 'break' statement");
      emitGoto(break_labels[break_level-1]);
      break;

    case Xcode.CONTINUE_STATEMENT: 
      if(continue_level <= 0) fatal("bad 'continue' statement");
      emitGoto(continue_labels[continue_level-1]);
      break;

      //
      //case Xcode.SWITCH_STATEMENT: 
      ///* find CASE_LABEL */
      //v = new Vector();
      //for(bp = b.getBody().getHead(); bp != null; bp = bp.getNext()){
      //if(bp.Opcode() == Xcode.CASE_LABEL){
      //v.add(new SwitchPair());
      //}
      //}
      //break;
		
    case Xcode.GOTO_STATEMENT:	/* (GOTO_STATEMENT label) */
      //	bb_next = (BasicBlock)labelBlocks.get(b.getLabel());
      // if(bb_next == null) fatal("GOTO label is not found");
      // bb.addCflowTo(bb_next);
	break;
	
    case Xcode.STATEMENT_LABEL:
    case Xcode.CASE_LABEL: 
    case Xcode.DEFAULT_LABEL: 

    case Xcode.RETURN_STATEMENT: /* (RETURN_STATEMENT value) */
      emitGoto(ret_label);
      break;

    default:
      fatal("unknown block:"+b);
    }
  }

  void emitBranch(BasicBlock bb,int label, boolean flag){
    emit(bb);
    emitLIR.emitBranchExpr(bb.getExpr(),label,flag);
  }

  void fatal(String msg){
    System.err.println("EmitInst:"+msg);
    System.exit(1);
  }
}


