import exc.object.*;

/**
 * Algorithm to emit LIR (Low-level Intermediate Lang) for expression.
 * This algorithm has two-pass. In the first pass, the order of code 
 * generation is estimated. In the second pass, actuall code is generated 
 * using previous information. 
 * In estimation pass, (1) check for call, (2) check logical expression
 * for branch
 * All register in LIR are pseudo-registers, which will be allocated into 
 * real hardware register in register-allocation phase.
 */

public class EmitLIR {

  EmitInstEnv env;	// link to EmitInstEnv

  // context value 
  final static int LABEL=0;	/* output label */
  final static int EXPR =1;	/* evaluate for side-effect */
  final static int CALL=2;	/* evaluate CALL */
  final static int ARG=3;	/* evaluate expression for argument */
  final static int BRANCH=4;	/* evaluate for conditional branch */
  final static int N_BRANCH=5;	/* evaluate for conditional branch */
  final static int GOTO=6;	/* output label */
  final static int CCODE=7;    	/* for 0-1 value after branch */

  final static int NOLAB = -1;

  final static int MAX_NODES = 100;
  int number_of_nodes;

  // these data structure are used for scheduling of emitting LIR.
  Xobject sched_nodes[] = new Xobject[MAX_NODES];
  int sched_context[] = new int[MAX_NODES];
  int sched_info[] = new int[MAX_NODES]; // label, number of args, args position
  int sched_count;

  // This table stores the register assigned to the node id.
  //  If the same register should be assigned to the nodes, 
  //  these have the same id.
  int node_regs[] = new int[MAX_NODES];

  public EmitLIR(EmitInstEnv env){
    this.env = env;
  }

  /** Emit LIR code for expression. */
  public void emitStatementExpr(Xobject e){
    // numbering Xobject with id
    numberingExpr(e);
    sched_count = 0;
    schedExpr(e);
    putSched(EXPR,e);
    doEmit();
  }

  /** Emit LIR code for branch. */
  public void emitBranchExpr(Xobject e,int label, boolean flag){
    numberingExpr(e);
    sched_count = 0;
    if(flag) schedBranch(e,NOLAB, label);
    else schedBranch(e,label,NOLAB);
    doEmit();
  }

  /* emit LIR according scheduling by pass 1 */
  void doEmit(){
    Operand opd;

    for(int i = 0; i < sched_count; i++){
      Xobject e = sched_nodes[i];
      int info = sched_info[i];
      switch(sched_context[i]){
      case LABEL:
	env.emit(Inst.Label(info));
	break;
      case GOTO:
	env.emit(Inst.Goto(info));
	break;
      case BRANCH:
	emitBranch(e,info,true);
	break;
      case N_BRANCH:
	emitBranch(e,info,false);
	break;
      case EXPR:
	emitExpr(e);
	break;

      case ARG:
	// opd = emitExpr(e);
      case CALL:

      default:
	fatal("unknown context");
      }
    }
  }

  void numberingExpr(Xobject x){
    XobjectIterator i = new bottomupXobjectIterator(x);
    int k = 0;
    for(i.init(); !i.end(); i.next()){
      i.getXobject().setId(k);
      node_regs[k] = -1;
    }
    k = number_of_nodes;
  }

  void putSched(int context,Xobject x, int info){
    sched_nodes[sched_count] = x;
    sched_context[sched_count] = context;
    sched_info[sched_count] = info;
    sched_count++;
  }

  void putSched(int context,Xobject x){
    sched_nodes[sched_count] = x;
    sched_context[sched_count] = context;
    sched_info[sched_count] = 0;
    sched_count++;
  }

  void putSched(int context,int info){
    sched_nodes[sched_count] = null;
    sched_context[sched_count] = context;
    sched_info[sched_count] = info;
    sched_count++;
  }


  // traverse Xobjects in top-down sched to decide the order of evaluation
  //  for expression. 
  void schedExpr(Xobject e){
    int n;

    switch(e.Opcode()){
    case Xcode.FUNCTION_CALL:
      schedExpr(e.operand());	/* eval function */
      n = schedArgs(e.right());
      putSched(CALL,e,n);
      return;
	
    case Xcode.COMMA_EXPR:
      /* these operation must be evaluated from left to right */
      /* ',' operator. must be executed sequentially */
      schedExpr(e.left());
      putSched(EXPR,e.left(),0);
      schedExpr(e);
      putSched(EXPR,e.right(),0);
      e.setId(e.right().Id());  // share register
      return;
	
    case Xcode.CONDITIONAL_EXPR:	/* ? and : */
      {
	int L1,L2,r;
	/*    comp&branch L1
	 *    r = exec(left)
	 *    jump L2
	 * L1:r = exec(right)
	 * L2:
	 */
	L1 = env.genLabel();
	L2 = env.genLabel();
	r = env.genRegN();
	schedBranch(e.left(),NOLAB,L1);
	schedExpr(e.right().left());
	putSched(EXPR,e.right().left(),r);
	putSched(GOTO,L2);
	putSched(LABEL,L1);
	schedExpr(e.right().right());
	putSched(EXPR,e.right().right(),r);
	putSched(LABEL,L2);	
	node_regs[e.Id()] = r;	/* set result register */
	return;
      }
    }

    if(e.isLogicalOp()){
      int L1;
      /* logical operators in the context of evaluting expression */
      /* if here, must be a logical operator for 0-1 value */
      /*     comp&branch L1 
       *     r = 1
       *     jump L2
       * L1: r = 0
       * L2:
       */
      L1 = env.genLabel();
      schedBranch(e, NOLAB,L1);
      putSched(CCODE,L1);
      return;
    }

    if(e.isAsgOp()){  /* for assignment, evaluate right and left */
	schedExpr(e.right());
	putSched(EXPR,e.right());
	schedExpr(e.left());
	return;
    }

    /* as default, evaluate left to right */
    for(XobjArgs args = e.getArgs(); args != null; args = args.nextArgs())
      schedExpr(args.getArg());
  }

  int schedArgs(Xobject args){
    fatal("no arg");
  }

  /* schedule branch code. */
  void schedBranch(Xobject e,int true_lab,int false_lab){
    int cxt,lab;

    switch(e.Opcode()){
    case Xcode.LOG_EQ_EXPR:
    case Xcode.LOG_NEQ_EXPR:
    case Xcode.LOG_GE_EXPR:
    case Xcode.LOG_GT_EXPR:
    case Xcode.LOG_LE_EXPR:
    case Xcode.LOG_LT_EXPR:
      schedExpr(e.left());
      schedExpr(e.right());
      cxt = BRANCH;
      if(true_lab == NOLAB){	// reverse condition 
	cxt = N_BRANCH;
	true_lab = false_lab;
	false_lab = NOLAB;
      }
      putSched(cxt,e,true_lab);
      if(false_lab != NOLAB) putSched(GOTO,null,false_lab);
      return;

    case Xcode.LOG_AND_EXPR:
      if(false_lab == NOLAB){
	lab = env.genLabel();
	schedBranch(e.left(),NOLAB,lab);
	schedBranch(e.right(),true_lab,false_lab);
	putSched(LABEL,null,lab);
      }  else {
	schedBranch(e.left(),NOLAB,false_lab);
	schedBranch(e.right(),true_lab,false_lab);
      }
      return;

    case Xcode.LOG_OR_EXPR:
      if(true_lab == NOLAB){
	lab = env.genLabel();
	schedBranch(e.left(),lab, NOLAB);
	schedBranch(e.right(),true_lab,false_lab);
	putSched(LABEL,lab);
      }  else {
	schedBranch(e.left(),false_lab,NOLAB);
	schedBranch(e.right(),true_lab,false_lab);
      }
      return;

    case Xcode.LOG_NOT_EXPR:
      schedBranch(e.operand(),false_lab,true_lab);
      return;
      
    case Xcode.COMMA_EXPR:
      schedExpr(e.left());
      putSched(EXPR,e.left());
      schedBranch(e.right(),true_lab,false_lab);
      return;

    default:
      schedExpr(e);
      if(true_lab == NOLAB)
	putSched(N_BRANCH,e,false_lab);
      else {
	putSched(BRANCH,e,true_lab);
	if(false_lab != NOLAB)
	  putSched(GOTO,false_lab);
      }
    }
  }

  /* emit LIR for branch */
  void emitBranch(Xobject e, int label, boolean rev_flag){
    Operand src1,src2;

    if(e.isLogicalOp()){
      if(e.isBinaryOp()){
	if(e.left().isIntConstant()){
	  src1 = Operand.Const(e.left());
	  src2 = emitExpr(e.right());
	} else if(e.right().isIntConstant()){
	  src1 = emitExpr(e.left());
	  src2 = Operand.Const(e.right());
	} else {
	  src1 = emitExpr(e.left());
	  src2 = emitExpr(e.right());
	}
	if(rev_flag) 
	  env.emit(Inst.BinaryIf(label,Xcons.revLogicalOpcode(e.Opcode()),
				 src1,src2));
	else
	  env.emit(Inst.BinaryIf(label,e.Opcode(),src2,src2));
	return;
      }
    }
    



  }

  /* emit LIR for expression */
  Operand emitExpr(Xobject e){
    Operand src1,src2;
    int dst;

    if(node_regs[e.Id()] != -1)  return Operand.Reg(dst);

    switch(e.Opcode()){
    case POINTER_REF:
      dst = env.genRegN();
      node_regs[e.Id()] = dst;
      env.emit(Inst.Load(dst,emitAddr(e.operand())));
      return Operand.Reg(dst);
      break;

    case Xcode.ASSIGN_EXPR:
      src1 = emitExpr(e.right());
      emitStore(e.left(),src1.getRegN());
      return src1;
    }

    if(e.AsgOp()){
      src1 = emitExpr(e.left());
      if(e.right().isIntConstant())
	src2 = Operand.Const(e.right());
      else
	src2 = emitExpr(e.left());
      dst = env.genRegN();
      node_regs[e.Id()] = dst;
      env.emit(Inst.Binary(Xcons.unAsgOpcode(e.Opcode()),dst,src1,src2));
      emitStore(e.left(),dst);
      return Operand.Reg(dst);
    }

    if(e.isBinaryOp()){
      if(e.left().isIntConstant()){
	src1 = Operand.Const(e.left());
	src2 = emitExpr(e.right());
      } else if(e.right().isIntConstant()){
	src1 = emitExpr(e.left());
	src2 = Operand.Const(e.right());
      } else {
	src1 = emitExpr(e.left());
	src2 = emitExpr(e.right());
      }
      dst = env.genRegN();
      node_regs[e.Id()] = dst;
      env.emit(Inst.Binary(e.Opcode(),dst,src1,src2));
      return Operand.Reg(dst);
    }

    if(e.isUnaryOp()){
      src1 = emitExpr(e.operand());
      dst = env.genRegN();
      node_regs[e.Id()] = dst;
      env.emit(Inst.Unary(e.Opcode(),dst,src1));
      return Operand.Reg(dst);
    }
  }

  void fatal(String msg){
    System.err.println("Fatal EmitLIR:"+msg);
    System.exit(1);
  }

}


