/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import exc.object.*;

/*
 * Data block for handling begin pragmas.
 *
 * A Translater object contains a linked list of ScmBlock objects
 * and calls process() for translation.
 */
public class ScmBlockOld implements Names {
    protected ScmBlockOld next;
    protected TranslaterOld translater;

    protected boolean invalid;
    protected String varName;
    protected Xtype varType;
    protected int dimension;
    protected Xobject[] sizeList;
    protected int[] strideList;

    public ScmBlockOld(Xobject name, Xobject params, TranslaterOld trans,
		    ScmBlockOld _next)
    {
	next = _next;
	translater = trans;
	invalid = false;
	varName = name.getSym();
	varType = translater.findCommonVariable(varName).Type();

	if (varType.getKind() != Xtype.ARRAY) {
	    translater.errorMessage(name, "not array");
	    invalid = true;
	    return;
	}

	dimension = varType.getNumDimensions();

	if (params.Nargs() != dimension * 2) {
	    translater.errorMessage(params, "bad parameters");
	    invalid = true;
	    return;
	}

        sizeList = new Xobject[dimension];
	strideList = new int[dimension];

	for (int i = 0; i < dimension; ++i) {
	    sizeList[i] = params.getArg(i);
	    Xobject stride = params.getArg(i + dimension);
	    if (stride instanceof XobjInt)
		strideList[i] = stride.getInt();
	    else {
		translater.errorMessage(params,
					"a stride width must be constant");
		invalid = true;
		return;
	    }
	}
    }

    public Xtype getType() { return varType; }

    public int getDimension() { return dimension; }

    public int getStride(int dim) { return strideList[dim]; }

    public ScmBlockOld lookup(String name) {
	ScmBlockOld b = this;
	while (b != null)
	    if (b.varName.equals(name))
		return b;
	    else
		b = b.next;

	return null;
    }

    public boolean invalidate(String name, boolean v) {
	ScmBlockOld b = lookup(name);
	if (b == null)
	    return false;
	else if (b.invalid)
	    return false;
	else {
	    b.invalid = v;
	    return true;
	}
    }

    /* Wraps the given program block with temporary variable declarations.
     *
     * Example: suppose the given block is:
     *
     * for (i = 0; i < N; ++i)
     *     x[i] = y[i + 1];
     *
     * Then, this method returns:
     *
     * { double* x_ptr;
     *   int x_offset_0; int x_size;
     *   x_ptr = scm_alloc(...);
     *   for (i = 0; i < N; ++i) {
     *       ?????????????;		translated by ScmBlock.process().
     *   }
     *   scm_free(x_ptr);
     * }
     *
     */
    public Xobject makeBlock(Xobject block) {
	StringBuffer code = new StringBuffer();
	code.append(makePointerType()).append(' ')
	    .append(getPointerName()).append(";\n");

	// System.out.println("DEBUG " + code);
	for (int i = 0; i < dimension; ++i)
	    code.append("int ").append(getOffsetName(i))
		.append(" = 0;\nint ").append(getCacheSize(i))
		.append(" = ").append(translater.toString(sizeList[i]))
		.append(";\n");
	// System.out.println("DEBUG " + code);

	extraVarDeclaration(code);
	makeAlloc(code);
	return new XobjList(Xcode.SCM_PRAGMA_CODE,
			    new XobjString(Xcode.IDENT,
					   "\n{\n" + code.toString()),
			    block,
			    new XobjString(Xcode.IDENT,
					   "\n" + makeFree() + "}\n"));
    }

    protected void extraVarDeclaration(StringBuffer code) {
	// do nothing.
    }

    protected void makeAlloc(StringBuffer code) {
	makeAlloc(getPointerName(), code);
    }

    protected void makeAlloc(String ptrName, StringBuffer code) {
	code.append(ptrName).append(" = " + ALLOC + "(sizeof(");
	code.append(makeElementType()).append(") * ");

	/* we don't need the original array size.

	Xtype t = getType();
	while (t.getKind() == Xtype.ARRAY) {
	    code += t.getArrayDim() + ", ";
	    t = t.getRef();
	}
	*/

	int k = 0;
	int dim = dimension;
	while (k < dim) {
	    String size = translater.toString(sizeList[k]);
	    size = "(" + size + ")"; //add by mstk
	    if (getStride(k) > 0)
		code.append("(").append(size).append(")/")
		    .append(getStride(k) + 1);
	    else
		code.append(size);

	    if (++k < dim)
		code.append(" * ");
	}

	code.append(");\n");
    }

    protected String makeFree() {
	return FREE + "(" + getPointerName() + ");\n";
    }

    // The following three methods return the names of temporary variables.

    public String getOffsetName(int dim) {
	return varName + "_offset" + dim;
    }

    public String getCacheSize(int dim) {
	return varName + "_size" + dim;
    }

    /* Returns the name of pointer to an on-chip memory area.
     */
    public String getPointerName() {
        // System.out.println("<DEBUG> getPointerName: " + varName + "_ptr");
	return varName + "_ptr";
    }

    /* Returns the type name of pointer to an array element.
     */
    public String makePointerType() {
	return makeElementType() + "*";
    }

    public String makeElementType() {
	Xtype t = varType;
	while (t.getKind() == Xtype.ARRAY)
	    t = t.getRef();

	String tname = t.toString();
	if (tname.equals("double") || tname.equals("int"))
	    return tname;
	else
	    return "struct _S0";	// the type should be complex.
					// this is very ad hoc!
    }

    /* Gets the C-language name of this array.
     */
    public String getArrayName() {
	Xobject decl = translater.findCommonVariable(varName).getValue();

	if (translater.findCommonVariable(varName).getStorageClass() == 
	    StorageClass.FCOMM){
	  // System.out.println("DEBUG " + decl);

	  Xobject commonName = decl.getArg(1).getArg(0);
	  Xobject memberName = decl.getArg(1).getArg(1);
	  // System.out.println("DEBUG " + commonName + memberName);
	  return commonName.getSym() + "._" + memberName.getInt()
	    + "." + varName;
	}
	else{
	  return varName;
	}
    }

    /*
     * Translates an expression for accessing an array on an on-chip cache.
     *
     * Example: supose that the original expression is:
     *
     *   x(i,j)
     *
     * then, this method translates it into:
     *
     *   x_ptr(i - x_offset0)(j - x_offset1)
     *
     * or
     *
     *   x_ptr((i - x_offset0)/<stride_width0>,
     *         (j - x_offset1)/<stride_width1>)
     *
     * Note: a Fortran array expression x(i,j) is represented by
     * x[j,i] in C.  (i and j are exchanged.)
     */
    public void process(Xobject node, TopdownIterator iterator) {
	Xobject xobj;

	if (invalid || !node.getSym().equals(varName)) {
	    if (next != null)
		next.process(node, iterator);

	    return;
	}

	// System.out.println("<DEBUG> getSym: " + node.getSym() + " " + varName);

	int dim = dimension;
	boolean first = true;
	Xobject[] index = new Xobject[dim];

	// System.out.println("DEBUG " + dim);

	while (--dim >= 0) {
	    iterator.moveUp();

	    // System.out.println("<DEBUG> ----------------------------- ");
	    if (first) {
		first = false;
		if (node.Opcode() == Xcode.IDENT){ 
		  Xobject p = iterator.getParent();
		  // System.out.println(p);
		  if (p.Opcode() != Xcode.PLUS_EXPR) {
		    processArrayName(iterator);
		    return;
		  }
		} else if (node.Opcode() == Xcode.FPARAM_ARRAY_ADDR){ 
		  Xobject p = iterator.get();
		  if (p.Opcode() != Xcode.PLUS_EXPR) {
		    processArrayName(iterator);
		    return;
		  }
		}
	    }

	    if (node.Opcode() == Xcode.IDENT)
	      iterator.moveUp();

	    xobj = new XobjString(Xcode.VAR, getOffsetName(dim));
	    // System.out.println(iterator.get().getArg(0).getArg(1));
	    xobj = new XobjList(Xcode.MINUS_EXPR,
				translater.process(iterator.get().getArg(1)),
				xobj);
	    // System.out.println(xobj);
	    index[dim] = xobj;
	}

	dim = dimension;
	xobj = index[dim - 1];
	for (int i = dim - 2; i >= 0; --i) {
	    xobj = new XobjList(Xcode.MUL_EXPR, xobj,
			new XobjString(Xcode.VAR, getCacheSize(i)));
	    xobj = new XobjList(Xcode.PLUS_EXPR, xobj, index[i]);

	    // if the stride width is more than zero, then the index is
	    // adjusted.

	    if (strideList[i] > 0)
		xobj = new XobjList(Xcode.DIV_EXPR, xobj,
				    Xcons.IntConstant(strideList[i] + 1));
	}

	xobj = new XobjList(Xcode.PLUS_EXPR,
			    new XobjString(Xcode.VAR, getPointerName()),
			    xobj);
	iterator.substitute(xobj);
	iterator.skipSubtree();
    }

    /*
     * Translates an array name without indexes, which may appear as
     * a subroutine parameter.  It is equivalent to <array name>(1,1,..).
     */
    protected void processArrayName(TopdownIterator iterator) {
	iterator.substitute(new XobjString(Xcode.VAR, getPointerName()));
	iterator.skipSubtree();
    }

//  protected static String getSymbol(Xobject xobj) {
//      return ((XobjString)xobj.getArg(0)).getSym();
//  }
}
