/*
 * Implementation note:
 *	grpc_function_handle_t and Ninf-G GridRPC API impelemtation
 *	seems not MT safe. Semantically, the handle seems NOT
 *	shareable between thread, since the handle is NOT a function
 *	abstraction, but a host abstraction, and degree of parallelism
 *	depends on number of the hosts.
 *
 *		m-hirano
 */

#include "ogr_hash.h"

#ifdef USE_PTHREAD
#include <pthread.h>

typedef pthread_mutex_t ogr_lock_t;
typedef pthread_cond_t ogr_cond_t;

#define ogr_init_lock(X)	pthread_mutex_init((X), NULL)
#define ogr_set_lock(X)		pthread_mutex_lock((X))
#define ogr_test_lock(X)	pthread_mutex_trylock((X))
#define ogr_unset_lock(X)	pthread_mutex_unlock((X))
#define ogr_cond_init(X)	pthread_cond_init((X), NULL)
#define ogr_cond_wait(X, Y)	pthread_cond_wait((X), (Y))
#define ogr_cond_broadcast(X)	pthread_cond_broadcast((X))
#define ogr_cond_signal(X)	pthread_cond_signal((X))

#else

typedef omp_lock_t ogr_lock_t;

#define ogr_init_lock(X)	omp_init_lock((X))
#define ogr_set_lock(X)		omp_set_lock((X))
#define ogr_test_lock(X)	omp_test_lock((X))
#define ogr_unset_lock(X)	omp_unset_lock((X))

#endif /* USE_PTHREAD */


typedef enum {
    ogr_HH_None = 0,
    ogr_HH_Allocating = 1,
    ogr_HH_Exists = 2,
    ogr_HH_Failed
} ogrHHStatus;


typedef enum {
    ogr_HandleIsFree = 0,	/* Free or just created */
    ogr_HandleIsAcquired = 1,	/* Free handle is just acquired to call */
    ogr_HandleIsCalling = 2,	/* RPC'ing rite now */
    ogr_HandleIsUsedAsync = 3,	/* A RPC invoked async */
    ogr_HandleIsAsyncAcquired = 4,
    				/* Async handle is just acquired to call */
    ogr_HandleIsWaiting	= 5,	/* Sync RPC or waiting 
				 * an async RPC */
    ogr_HandleIsUnusable	/* handle allocation failed */
} ogrHandleUsage;


typedef struct {
    struct timeval start;
    struct timeval end;
    double MREtime;
    int nCalls;
    ogrHandleUsage type;
    ogr_lock_t usageLock;
} ogrHostHandleUsage;


struct ogrFuncHandleRec;

struct ogrHostHandleRec {
    char *hostName;
			/* hostname of handle, malloc'd */
    char *funcName;
			/* funcname of handle, malloc'd */
    grpc_function_handle_t *gfhPtr;
    			/* function handle POINTER, malloc'd
			   allocated by ogrRpcHostInfo */
    ogr_lock_t hostLock;
    grpc_sessionid_t sId;
			/* current session running on hPtr */
    struct ogrFuncHandleRec *ofhPtr;
			/* parent function handle POINTER */
    int idx;		/* index of ofhPtr's hHandleList */
};
typedef struct ogrHostHandleRec ogrHostHandle;

struct ogrFuncHandleRec {
    char *funcName;	/* RPC module/function name */
    int nMaxHandles;	/* Max number of handles (== # of hosts) */
    int nCurHandles;	/* Cur number of handles (<= nMaxHandles) */
    ogrHostHandle **hHandleList;
    			/* per-host function handle, malloc'd */
    ogrHostHandleUsage *hUsageList;
    			/* host handle usage. Access with locking this
			 * func handle */
    ogr_lock_t funcLock;

    ogr_lock_t fhInitLock;
    ogr_cond_t fhInitCond;

    ogr_lock_t hhWaitLock;
    ogr_cond_t hhWaitCond;

    int MRUHandle;
};
typedef struct ogrFuncHandleRec ogrFuncHandle;

typedef struct {
    char *hostName;	/* RPC host name */
    char *funcName;	/* RPC module/fuction name */
    int nHandles;
    grpc_function_handle_t *gfhPtr;
    			/* function handle array, malloc'd */
    ogr_lock_t funcInfoLock;
    int rpcFuncHandleInited;
    int rpcFuncHandleDeleted;
} ogrRpcFuncInfo;

typedef struct {
    char *hostName;
    int nHandles;
    ogr_HashTable hostFuncTable;
    ogr_lock_t hostInfoLock;
} ogrRpcHostInfo;

#if 0
#define DOUT_H(X, Y)	fprintf(stderr, "lock debug: host '%s:%d' %s\n", (X)->hostName, (X)->idx, Y)
#define DOUT_F(X, Y)	fprintf(stderr, "lock debug: func '%s' %s\n", (X)->funcName, Y)
#else
#define DOUT_H(X, Y) { ; }
#define DOUT_F(X, Y) { ; }
#endif

#define ogrFuncOfHost(X)	((X)->ofhPtr)

#define ogrFunc2HostUsage(X, idx)	\
	((X)->hUsageList[idx])

#define ogrFunc2HostUsageType(X, idx)	\
	(ogrFunc2HostUsage((X), (idx))).type
#define ogrFunc2HostUsageLock(X, idx)	\
	(ogrFunc2HostUsage((X), (idx))).usageLock
#define ogrFunc2HostUsageNCalls(X, idx)	\
	(ogrFunc2HostUsage((X), (idx))).nCalls

#define ogrLockHostUsageFromFunc(X, idx) \
	ogr_set_lock(&(ogrFunc2HostUsageLock((X), (idx))))

#define ogrUnlockHostUsageFromFunc(X, idx) \
	ogr_unset_lock(&(ogrFunc2HostUsageLock((X), (idx))))

#define ogrHostUsage(X)	\
	ogrFunc2HostUsage((X)->ofhPtr, (X)->idx)

#define ogrHostUsageType(X)	\
	ogrFunc2HostUsageType((X)->ofhPtr, (X)->idx)
#define ogrHostUsageLock(X)	\
	ogrFunc2HostUsageLock((X)->ofhPtr, (X)->idx)
#define ogrHostUsageNCalls(X)	\
	ogrFunc2HostUsageNCalls((X)->ofhPtr, (X)->idx)

#define ogrLockHostUsage(X)	ogr_set_lock(&(ogrHostUsageLock((X))))
#define ogrUnlockHostUsage(X)	ogr_unset_lock(&(ogrHostUsageLock((X))))

#define ogrUpdateHostUsage(X, st)	\
{	\
	ogrLockHostUsage((X)); \
	ogrHostUsageType((X)) = (st); \
	ogrUnlockHostUsage((X)); \
}

#define ogrIncrHostUsageNCalls(X)	\
{	\
	ogrLockHostUsage((X)); \
	ogrHostUsageNCalls((X))++; \
	ogrUnlockHostUsage((X)); \
}

#define ogrLockHost(X)		{ ogr_set_lock(&((X)->hostLock)); DOUT_H((X), "lock"); }
#define ogrUnlockHost(X)	{ ogr_unset_lock(&((X)->hostLock)); DOUT_H((X), "unlock"); }

#define ogrLockFunc(X)		{ ogr_set_lock(&((X)->funcLock)); DOUT_F((X), "lock"); }
#define ogrUnlockFunc(X)	{ ogr_unset_lock(&((X)->funcLock)); DOUT_F((X), "unlock"); }

#define ogrLockFHInit(X)	ogr_set_lock(&((X)->fhInitLock))
#define ogrUnlockFHInit(X)	ogr_unset_lock(&((X)->fhInitLock))
#define ogrCondSleepFHInit(X)	ogr_cond_wait(&((X)->fhInitCond), &((X)->fhInitLock))
#define ogrCondWakeupFHInit(X)	ogr_cond_broadcast(&((X)->fhInitCond))

#define ogrLockHHWait(X)	ogr_set_lock(&((X)->hhWaitLock))
#define ogrUnlockHHWait(X)	ogr_unset_lock(&((X)->hhWaitLock))
#define ogrCondSleepHHWait(X)	ogr_cond_wait(&((X)->hhWaitCond), &((X)->hhWaitLock))
#define ogrCondWakeupHHWait(X)	ogr_cond_broadcast(&((X)->hhWaitCond))

#define ogrLockFuncInfo(X)	ogr_set_lock(&((X)->funcInfoLock))
#define ogrUnlockFuncInfo(X)	ogr_unset_lock(&((X)->funcInfoLock))
#define ogrTryLockFuncInfo(X)	ogr_test_lock(&((X)->funcInfoLock))

#define ogrLockHostInfo(X)	ogr_set_lock(&((X)->hostInfoLock))
#define ogrUnlockHostInfo(X)	ogr_unset_lock(&((X)->hostInfoLock))
#define ogrTryLockHostInfo(X)	ogr_test_lock(&((X)->hostInfoLock))

extern void	_ogr_debug_init(void);
extern void	_ogr_debug_final(void);

extern void	_ogr_debug(int lvl, ...);
extern void	_ogr_warning(char *fmt, ...);
extern void	_ogr_error(char *fmt, ...);

extern void	_ogr_stat_init(void);
extern void	_ogr_stat_final(void);

extern void	_ogr_start_timer(char *id);
extern void	_ogr_stop_timer(char *id);
extern double	_ogr_get_timer(char *id);

extern int	_ogr_init(int argc, char *argv[]);
extern void	_ogr_final(void);

extern void	_ogr_rpc_prologue(void);

extern ogrHostHandle *	_ogr_getHostHandle(char *funcName);

extern void		_ogr_waitAll(void);

extern int		_ogr_barrierVar_init(void);

extern int		_ogr_create_barrier(char *bVar);
extern int		_ogr_reset_barrier(char *bVar);
extern int		_ogr_append_session_to_barrier(char *bVar,
						       grpc_sessionid_t sId);
extern grpc_sessionid_t *	_ogr_get_session_list_from_barrier(char *bVar,
								   int *nPtr);

