static char rcsid[] = "$Id: ogr_misc.c,v 1.2 2004-10-21 08:34:53 m-hirano Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
#include <float.h>
#include "exc_platform.h"
#include "grpc.h"
#include "ogr_hash.h"
#include "omp.h"
#include "ogr_internal.h"


typedef struct {
    char *id;
    struct timeval start;
    int isStart;
    struct timeval stop;
    int isStop;
    ogr_lock_t lock;
} ogrTimer;
#define ogrLockTimer(X)		ogr_set_lock(&((X)->lock))
#define ogrUnlockTimer(X)	ogr_unset_lock(&((X)->lock))


static ogr_lock_t debugLock;

static int _ogr_debugLvl = 0;
static FILE *_ogr_debugFd = NULL;

static ogr_HashTable timerTable;
static ogr_lock_t timerLock;
static int _ogr_doStat = 0;


static char *
getDate(char *buf) {
    time_t x = time(NULL);
    char *fmt = "%T(%m/%d/%Y)";
    strftime(buf, 1024, fmt, localtime(&x));
    return buf;
}


void
_ogr_debug_init(void) {
    char *envDebugLvl = getenv("OGR_DEBUG_LVL");

    if (envDebugLvl != NULL && envDebugLvl[0] != '\0') {
	int tmpDl = atoi(envDebugLvl);

	if (tmpDl > 0) {
	    _ogr_debugLvl = tmpDl;
	}

	if (_ogr_debugLvl > 0) {
	    FILE *tmpFd = NULL;
	    char *envDebugFd = getenv("OGR_DEBUG_FILE");

	    if (envDebugFd != NULL && envDebugFd[0] != '\0') {
		tmpFd = fopen(envDebugFd, "a+");
		if (tmpFd != NULL) {
		    _ogr_debugFd = tmpFd;
		} else {
		    fprintf(stderr, "OpenGR runtime: "
			    "can't open %s for debug. "
			    "use stderr instead.\n",
			    envDebugFd);
		    _ogr_debugFd = stderr;
		}
	    } else {
		_ogr_debugFd = stderr;
	    }

	    ogr_init_lock(&debugLock);
	}
    }
}


void
_ogr_debug_final(void) {
    if (_ogr_debugFd != NULL &&
	_ogr_debugFd != stderr) {

	ogr_set_lock(&debugLock);

	fflush(_ogr_debugFd);
	fclose(_ogr_debugFd);

	_ogr_debugFd = NULL;

	ogr_unset_lock(&debugLock);
    }
}


void
_ogr_error(char *fmt, ...) {
    va_list args;
    int tid = omp_get_thread_num();
    char buf[1024];

    ogr_set_lock(&debugLock);

    fprintf(stderr, "%s@tid=%03d: error: ", getDate(buf), tid);

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);

    ogr_unset_lock(&debugLock);
}


void
_ogr_warning(char *fmt, ...) {
    va_list args;
    int tid = omp_get_thread_num();
    char buf[1024];

    ogr_set_lock(&debugLock);

    fprintf(stderr, "%s@tid=%03d: warning: ", getDate(buf), tid);

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);

    ogr_unset_lock(&debugLock);
}


void
_ogr_debug(int lvl, ...) {
    if (_ogr_debugFd != NULL &&
	lvl <= _ogr_debugLvl) {
	va_list args;
	char *fmt;
	char buf[1024];

	int tid = omp_get_thread_num();

	va_start(args, lvl);
	fmt = va_arg(args, char *);

	ogr_set_lock(&debugLock);

	fprintf(_ogr_debugFd, "%s@tid=%03d: ",
		    getDate(buf), tid);
	vfprintf(_ogr_debugFd, fmt, args);
	fflush(_ogr_debugFd);

	ogr_unset_lock(&debugLock);

	va_end(args);
    }
}


static ogrTimer *
allocTimer(char *id) {
    ogrTimer *ret = (ogrTimer *)malloc(sizeof(ogrTimer));

    if (ret == NULL) {
	_ogr_error("can't allocate timer object '%s'.\n");
	abort();
    }

    (void)memset((void *)ret, 0, sizeof(ogrTimer));
    ret->id = strdup(id);
    ogr_init_lock(&(ret->lock));

    return ret;
}


static void
startTimer(ogrTimer *tPtr) {
    ogrLockTimer(tPtr);
    gettimeofday(&(tPtr->start), NULL);
    tPtr->isStart = 1;
    ogrUnlockTimer(tPtr);
}


static void
stopTimer(ogrTimer *tPtr) {
    ogrLockTimer(tPtr);
    gettimeofday(&(tPtr->stop), NULL);
    tPtr->isStop = 1;
    ogrUnlockTimer(tPtr);
}


static double
calcTimer(ogrTimer *tPtr) {
    double ret;
    ogrLockTimer(tPtr);
    if (tPtr->isStart != 1) {
	gettimeofday(&(tPtr->start), NULL);
    }
    if (tPtr->isStop != 1) {
	gettimeofday(&(tPtr->stop), NULL);
    }
    ret = ((tPtr->stop.tv_sec * 1000000 + tPtr->stop.tv_usec) -
	   (tPtr->start.tv_sec * 1000000 + tPtr->start.tv_usec)) /
	1000000.0;
    ogrUnlockTimer(tPtr);

    return ret;
}


static ogrTimer *
getTimer(char *id) {
    ogr_HashEntry *ePtr = NULL;
    ogrTimer *ret = NULL;
    int isNew = 0;

    ePtr = ogr_CreateHashEntry(&timerTable, id, &isNew);
    if (isNew == 1) {
	ret = allocTimer(id);
	ogr_SetHashValue(ePtr, (ClientData)ret);
    } else {
	ret = (ogrTimer *)ogr_GetHashValue(ePtr);
    }

    return ret;
}


void
_ogr_start_timer(char *id) {
    if (_ogr_doStat == 1) {
	ogrTimer *tPtr = NULL;    

	ogr_set_lock(&timerLock);
	tPtr = getTimer(id);
	ogr_unset_lock(&timerLock);

	if (tPtr != NULL) {
	    startTimer(tPtr);
	}
    }
}


void
_ogr_stop_timer(char *id) {
    if (_ogr_doStat == 1) {
	ogrTimer *tPtr = NULL;   

	ogr_set_lock(&timerLock);
	tPtr = getTimer(id);
	ogr_unset_lock(&timerLock);

	if (tPtr != NULL) {
	    stopTimer(tPtr);
	}
    }
}


double
_ogr_get_timer(char *id) {
    if (_ogr_doStat == 1) {
	ogrTimer *tPtr = NULL;   

	ogr_set_lock(&timerLock);
	tPtr = getTimer(id);
	ogr_unset_lock(&timerLock);

	if (tPtr != NULL) {
	    return calcTimer(tPtr);
	} else {
	    return DBL_MAX;
	}
    } else {
	return DBL_MAX;
    }	
}


void
_ogr_stat_init(void) {
    char *envDoStat = getenv("OGR_STAT");

    if (envDoStat != NULL) {
	if (strcasecmp(envDoStat, "1") == 0 ||
	    strcasecmp(envDoStat, "Yes") == 0 ||
	    strcasecmp(envDoStat, "True") == 0) {
	    _ogr_doStat = 1;

	    ogr_init_lock(&timerLock);
	    ogr_InitHashTable(&timerTable, OGR_STRING_KEYS);
	}
    }
}


#define FOR_ENTRIES_IN_HASH(h, s, tab) \
	for (h = ogr_FirstHashEntry(tab, s); \
	     h != NULL; \
	     h = ogr_NextHashEntry(s))

void
_ogr_stat_final(void) {
    if (_ogr_doStat == 1) {
	extern ogr_HashTable funcTable;
	ogr_HashEntry *ePtr;
	ogr_HashSearch sctx;
	ogrTimer *tPtr;
	FILE *fd = stderr;
	ogrFuncHandle *fhPtr;
	int i;
	int nCalls = 0;

	if (_ogr_debugLvl > 0) {
	    if (_ogr_debugFd != NULL) {
		fd = _ogr_debugFd;
	    }
	}

	fprintf(fd, "\nStatistics:\n\n");

	FOR_ENTRIES_IN_HASH(ePtr, &sctx, &timerTable) {
	    tPtr = (ogrTimer *)ogr_GetHashValue(ePtr);
	    fprintf(fd, "timer '%s':\t%f sec.\n",
		    tPtr->id, calcTimer(tPtr));
	}

	fprintf(fd, "\nHandle usage:\n\n");

	FOR_ENTRIES_IN_HASH(ePtr, &sctx, &funcTable) {
	    fhPtr = (ogrFuncHandle *)ogr_GetHashValue(ePtr);
	    for (i = 0; i < fhPtr->nMaxHandles; i++) {
		fprintf(fd, "[%3d] %s@%s: %d\n",
			i,
			fhPtr->hHandleList[i]->funcName,
			fhPtr->hHandleList[i]->hostName,
			fhPtr->hUsageList[i].nCalls);
		nCalls += fhPtr->hUsageList[i].nCalls;
	    }
	}
	fprintf(fd, "\nTotal calls: %d\n\n", nCalls);
    }
}
