/*
 *  $Id: hubbard.c,v 1.1.1.1 2004-02-06 18:15:00 msato Exp $
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include <tdl.h>
#include "ttime.h"

/*
 *  Input file is assumed to be sorted and stored in COO format.
 */

static FILE *
open_file(char *filename)
{
    FILE *fd = fopen(filename, "r");
    if (fd == NULL) {
	perror("fopen");
	exit(1);
    }
    return fd;
}

static int
close_file(FILE *fd)
{
    return fclose(fd);
}

static int
read_size(FILE *fd, int *mat_size, int *nzero)
{
    /**  Get matrix size and num of nonzero elements.  **/

    return fscanf(fd, "%d %d", mat_size, nzero);
}

static void
read_mat(FILE *fd, int size, int nzero, int *idx, int *ib, double *a)
{
    int d1, d2;
    double el;
    int prev_d1 = -1;
    int k = 0, l = 0;

    while (fscanf(fd, "%d %d %lf", &d1, &d2, &el) != EOF) {
	if (prev_d1 != d1) {
	    ib[l] = k;
	    ++l;
	}
	idx[k] = d2 - 1;  /**  zero origin  **/
	a[k] = el;

	++k;
	prev_d1 = d1;
    }
    ib[l] = k;

    assert(size == l);
    assert(nzero == k);
}

/*
 *  MAIN
 */

/* #define TEST */

#define NUM_ITER	3

int *idx, *ib;
TDL_desc idx_desc, ib_desc;
double *a;
TDL_desc a_desc;
double *b, *c;
TDL_desc b_desc, c_desc;

extern void
mult(int n, int *idx, int *ib,
     double *a, double *b, double *c,
     TDL_desc idx_desc, TDL_desc ib_desc,
     TDL_desc a_desc, TDL_desc b_desc, TDL_desc c_desc);

int
main(int argc, char *argv[])
{
    int size, nzero;
    int i;
    ttime_t tt1, tt2;
    FILE *fd;
#ifdef RHiNET
    GID gid = 0x200;
    int thread_num, num_threads, daemonport;
    char *daemonhost;
#endif

    /**  Initialization  **/

#ifdef MPI
    TDL_init(&argc, &argv);

    if (argc != 2)
	printf("usage: %s filename\n", argv[0]), exit(1);
#elif defined(RHiNET)
    if (argc != 6)
	printf("usage: %s filename thread_num num_threads daemonhost daemonport\n", argv[0]), exit(1);

    thread_num = atoi(argv[2]);
    num_threads = atoi(argv[3]);

    daemonhost = argv[4];
    daemonport = atoi(argv[5]);

    TDL_init(&argc, &argv,
	     num_threads, thread_num, daemonhost, daemonport);
#else
#error Not support the message library
#endif

    fd = open_file(argv[1]);
    read_size(fd, &size, &nzero);

if (TDL_gThreadNum == 0)
    printf("size = %d, nzero = %d\n", size, nzero);

#ifdef MPI
    idx = (int*)malloc(sizeof(int) * nzero);
    idx_desc = TDL_desc_create(idx, sizeof(int) * nzero, TDL_MASTER);
    ib = (int*)malloc(sizeof(int) * (size + 1));
    ib_desc = TDL_desc_create(ib, sizeof(int) * (size + 1), TDL_MASTER);
    a = (double*)malloc(sizeof(double) * nzero);
    a_desc = TDL_desc_create(a, sizeof(double) * nzero, TDL_MASTER);
#elif defined(RHiNET)
    idx = TDL_malloc(sizeof(int) * nzero);
    idx_desc = TDL_desc_create(idx, sizeof(int) * nzero, TDL_MASTER, gid);
    gid += TDL_NUM_PAGES_FROM_SIZE(sizeof(int) * nzero);
    ib = TDL_malloc(sizeof(int) * (size + 1));
    ib_desc = TDL_desc_create(ib, sizeof(int) * (size + 1), TDL_MASTER, gid);
    gid += TDL_NUM_PAGES_FROM_SIZE(sizeof(int) * (size + 1));
    a = TDL_malloc(sizeof(double) * nzero);
    a_desc = TDL_desc_create(a, sizeof(double) * nzero, TDL_MASTER, gid);
    gid += TDL_NUM_PAGES_FROM_SIZE(sizeof(double) * nzero);
#endif

    if (TDL_gThreadNum == 0) {
	tgettime(tt1);
	read_mat(fd, size, nzero, idx, ib, a);
	tgettime(tt2);
	printf("read matrix: done (%g sec)\n", ttime_diff_sec(tt1, tt2));
    }
    close_file(fd);

    /**  Change distribution  **/

    {
	TDL_desc idx_desc0, ib_desc0, a_desc0;

	idx_desc0 = TDL_desc_create_desc(idx_desc);
	ib_desc0 = TDL_desc_create_desc(ib_desc);
	a_desc0 = TDL_desc_create_desc(a_desc);

	TDL_desc_dist(idx_desc0, TDL_BLOCK);
	TDL_desc_dist(ib_desc0, TDL_BLOCK);
	TDL_desc_dist(a_desc0, TDL_BLOCK);

	TDL_fence();
	TDL_remap(idx_desc, idx_desc0);
	TDL_remap(ib_desc, ib_desc0);
	TDL_remap(a_desc, a_desc0);
	TDL_fence();

	TDL_desc_free(idx_desc0);
	TDL_desc_free(ib_desc0);
	TDL_desc_free(a_desc0);
    }

    /**  Setup vector. **/

#ifdef MPI
    b = (double*)malloc(sizeof(double) * size);
    b_desc = TDL_desc_create(b, sizeof(double) * size, TDL_BLOCK);
    c = (double*)malloc(sizeof(double) * size);
    c_desc = TDL_desc_create(c, sizeof(double) * size, TDL_BLOCK);
#elif defined(RHiNET)
    b = TDL_malloc(sizeof(double) * size);
    b_desc = TDL_desc_create(b, sizeof(double) * size, TDL_BLOCK, gid);
    gid += TDL_NUM_PAGES_FROM_SIZE(sizeof(double) * size);
    c = TDL_malloc(sizeof(double) * size);
    c_desc = TDL_desc_create(c, sizeof(double) * size, TDL_BLOCK, gid);
    gid += TDL_NUM_PAGES_FROM_SIZE(sizeof(double) * size);
#endif

    for (i = 0; i < size; ++i)
	if (TDL_IS_VALID(b_desc, &b[i]))
	    b[i] = 1;

    for (i = 0; i < size; ++i)
	if (TDL_IS_VALID(c_desc, &c[i]))
	    c[i] = 0;

    /**  MAIN  **/

    mult(size, idx, ib, a, b, c,
	 idx_desc, ib_desc, a_desc, b_desc, c_desc);

    /**  Finalization  **/

/*    free((void*)idx); */
    TDL_desc_free(idx_desc);
/*    free((void*)ib); */
    TDL_desc_free(ib_desc);
/*    free((void*)a); */
    TDL_desc_free(a_desc);
/*    free((void*)b); */
    TDL_desc_free(b_desc);
/*    free((void*)c); */
    TDL_desc_free(c_desc);

    TDL_finalize();

    exit(0);
}


