/*
 *  Matrix-vector multiplication
 *
 *  the array 'c' is inspected in the first step.
 *
 *  $Id: mult4.c,v 1.1.1.1 2004-02-06 18:15:01 msato Exp $
 */

/*
 *  Compressed Sparse Row Format
 *
 *  inputs:
 *
 *  n     --- matrix size		(int)
 *  idx   --- index array of size n     (int*)
 *  ib    --- arrays indicating beginning/last position of i'th row
 *            of 'a'                    (int*)
 *  a     --- nonzero elements          (double*)
 *  b     --- vector                    (double*)
 *
 *  output:
 *
 *  c     --- vector (= c + a * b)      (double*)
 */

void
vec_mult_CSR(n, idx, ib, a, b, c)
    int n;
    int *idx, *ib;
    double *a, *b, *c;
{
    int i, j;

    for (i = 0; i < n; ++i) {
	double t = 0;
	for (j = ib[i]; j < ib[i + 1]; ++j)
	    t += a[j] * b[idx[j]];
	c[i] += t;
    }
}

/*
 *
 */

/* #define TIME_ONLY_COMPUTATION */
/* #define OUTPUT_LOOP_INFO */
/* #define INPUT_LOOP_INFO */
#define FILENAME	"vec_mult_desc"

#define NO_MERGE

#include <stdio.h>
#include <stdlib.h>

#ifdef TIME_ONLY_COMPUTATION
#include "ttime.h"
#endif

#include <tdl.h>

void
vec_mult_CSR_IE(n, idx, ib, a, b, c
#ifdef TIME_ONLY_COMPUTATION
		, time
#endif
		, idx_desc, ib_desc, a_desc, b_desc, c_desc)
    int n;
    int *idx, *ib;
    double *a, *b, *c;
#ifdef TIME_ONLY_COMPUTATION
    double *time;
#endif
    TDL_desc idx_desc, ib_desc, a_desc, b_desc, c_desc;
{
    static int n0 = 0;
    static TDL_desc idx_desc0, ib_desc0;
    static TDL_desc a_desc0, b_desc0, c_desc0;
    int i, j;

#if 0   /**  original code  **/

    for (i = 0; i < n; ++i) {
	double t = 0;
	for (j = ib[i]; j < ib[i + 1]; ++j)
	    t += a[j] * b[idx[j]];
	c[i] += t;
    }
#endif

    {
	int start, end, step;
	int l_start, l_end, l_step;
    
	start = 0;
	end = n;
	step = 1;

	/**  Compute area of Computation  **/

	TDL_loop_schedule(TDL_BLOCK, start, end, step,
			  &l_start, &l_end, &l_step);

	if (n0 == 0 || n != n0
	    || idx_desc0 == (TDL_desc)NULL || ib_desc0 == (TDL_desc)NULL
	    || a_desc0 == (TDL_desc)NULL || b_desc0 == (TDL_desc)NULL
	    || c_desc0 == (TDL_desc)NULL
	    || TDL_BASE(idx_desc) != TDL_BASE(idx_desc0)
	    || TDL_BASE(ib_desc) != TDL_BASE(ib_desc0)) {

	    /*
	     *  Phase I: Inspector
	     */

	    n0 = n;

	    /**  Create new descriptors of distributed arrays.  **/

	    if (idx_desc0 != NULL)
		TDL_desc_free(idx_desc0);
	    if (ib_desc0 != NULL)
		TDL_desc_free(ib_desc0);
	    if (a_desc0 != NULL)
		TDL_desc_free(a_desc0);
	    if (b_desc0 != NULL)
		TDL_desc_free(b_desc0);
	    if (c_desc0 != NULL)
		TDL_desc_free(c_desc0);

	    idx_desc0 = TDL_desc_create_desc(idx_desc);
	    ib_desc0 = TDL_desc_create_desc(ib_desc);
	    a_desc0 = TDL_desc_create_desc(a_desc);
	    b_desc0 = TDL_desc_create_desc(b_desc);
	    c_desc0 = TDL_desc_create_desc(c_desc);

#ifndef INPUT_LOOP_INFO

	    /**  Inspect access pattern (1) **/

	    for (i = l_start; i < l_end; i += l_step) {
		TDL_inspect_home((void*)&ib[i], sizeof(int), ib_desc0);
		TDL_inspect_home((void*)&ib[i + 1], sizeof(int), ib_desc0);
		TDL_inspect_home((void*)&c[i], sizeof(double), c_desc0);
	    }

	    /**  Decide distribution of ib[] and c[]  **/

	    TDL_desc_fix_home(ib_desc0);
	    TDL_desc_fix_home(c_desc0);

	    /*
	     *  If there is a small difference, merge sharing information.
	     */
#ifndef NO_MERGE
	    TDL_desc_merge(ib_desc, ib_desc0);
	    TDL_desc_merge(c_desc, c_desc0);
#endif
	    /**  Inspect access pattern (2) **/

	    for (i = l_start; i < l_end; i += l_step) {
		TDL_inspect_ref((void*)&ib[i], sizeof(int), ib_desc0);
		TDL_inspect_ref((void*)&ib[i + 1], sizeof(int), ib_desc0);
		TDL_inspect_ref((void*)&c[i], sizeof(double), c_desc0);
		TDL_inspect_modify((void*)&c[i], sizeof(double), c_desc0);
	    }

	    /**  Message aggregation & creation of send list. **/

	    TDL_desc_fix_ref(ib_desc0);
	    TDL_desc_fix_ref(c_desc0);

	    /**  Message aggregation & creation of recv list. **/

	    TDL_desc_fix_modify(c_desc0);

	    /**  Redistribute & change current distribution.  **/

	    TDL_fence();
	    TDL_remap_ref(ib_desc, ib_desc0);
	    TDL_fence();

	    /**  Inspect access pattern (3) **/

	    for (i = l_start; i < l_end; i += l_step) {
		for (j = ib[i]; j < ib[i + 1]; ++j) {
		    TDL_inspect_home((void*)&a[j], sizeof(double), a_desc0);
		    TDL_inspect_home((void*)&idx[j], sizeof(int), idx_desc0);
		}
	    }

	    /**  Decide distribution of a[], idx[]  **/

	    TDL_desc_fix_home(a_desc0);
	    TDL_desc_fix_home(idx_desc0);

	    /*
	     *  If there is a small difference, merge sharing information.
	     */
#ifndef NO_MERGE
	    TDL_desc_merge(a_desc, a_desc0);
	    TDL_desc_merge(idx_desc, idx_desc0);
#endif
	    /**  Inspect access pattern (4) **/

	    for (i = l_start; i < l_end; i += l_step) {
		for (j = ib[i]; j < ib[i + 1]; ++j) {
		    TDL_inspect_ref((void*)&a[j], sizeof(double), a_desc0);
		    TDL_inspect_ref((void*)&idx[j], sizeof(int), idx_desc0);
		}
	    }

	    /**  Message aggregation & creation of send list. **/

	    TDL_desc_fix_ref(a_desc0);
	    TDL_desc_fix_ref(idx_desc0);

	    /**  Redistribution  **/

	    TDL_remap_ref(idx_desc, idx_desc0);
	    TDL_fence();

	    /**  Inspect access pattern (5) **/

	    for (i = l_start; i < l_end; i += l_step) {
		for (j = ib[i]; j < ib[i + 1]; ++j) {
		    TDL_inspect_home((void*)&b[idx[j]], sizeof(double), b_desc0);
		}
	    }

	    /**  Decide distribution of b[]  **/

	    TDL_desc_fix_home(b_desc0);

	    /*
	     *  If there is a small difference, merge sharing information.
	     */
#ifndef NO_MERGE
	    TDL_desc_merge(b_desc, b_desc0);
#endif
	    /**  Inspect access pattern (6) **/

	    for (i = l_start; i < l_end; i += l_step) {
		for (j = ib[i]; j < ib[i + 1]; ++j) {
		    TDL_inspect_ref((void*)&b[idx[j]], sizeof(double), 
				    b_desc0);
		}
	    }

	    /**  Message aggregation & creation of send list. **/

	    TDL_desc_fix_ref(b_desc0);

#else  /* INPUT_LOOP_INFO */
	    {
		char fn[255];
		FILE *fd;

		sprintf(fn, "%s-%d", FILENAME, TDL_gThreadNum);
		if ((fd = fopen(fn, "r")) == NULL)
		    perror("fopen"), exit(1);
		TDL_desc_read(fd, idx_desc0);
		TDL_desc_read(fd, ib_desc0);
		TDL_desc_read(fd, a_desc0);
		TDL_desc_read(fd, b_desc0);
		TDL_desc_read(fd, c_desc0);
		fclose(fd);

		/**  Create send/recv list.  **/

		TDL_desc_fix_ref(idx_desc0);
		TDL_desc_fix_ref(ib_desc0);
		TDL_desc_fix_ref(a_desc0);
		TDL_desc_fix_ref(b_desc0);
		TDL_desc_fix_ref(c_desc0);
		TDL_desc_fix_modify(c_desc0);
	    }

#endif  /* INPUT_LOOP_INFO */

#ifdef OUTPUT_LOOP_INFO
	    {
		char fn[256];
		FILE *fd;
		sprintf(fn, "%s-%d", FILENAME, TDL_gThreadNum);
		if ((fd = fopen(fn, "w")) == NULL)
		    perror("fopen"), exit(1);
		TDL_desc_write(fd, idx_desc0);
		TDL_desc_write(fd, ib_desc0);
		TDL_desc_write(fd, a_desc0);
		TDL_desc_write(fd, b_desc0);
		TDL_desc_write(fd, c_desc0);
		fclose(fd);
	    }
#endif

#if 0
if (TDL_gThreadNum == 0) {
    puts("idx:");
    TDL_desc_print_abs(idx_desc0);
    puts("ib:");
    TDL_desc_print_abs(ib_desc0);
    puts("a:");
    TDL_desc_print_abs(a_desc0);
    puts("b:");
    TDL_desc_print(b_desc0);
    puts("c:");
    TDL_desc_print_abs(c_desc0);
}
#endif

	}

	/*
	 *  Phase II: Executor
	 */

	/**  Redistribute & change current distributions.  **/

	TDL_fence();
	TDL_remap_ref(idx_desc, idx_desc0);
	TDL_remap_ref(ib_desc, ib_desc0);
	TDL_remap_ref(a_desc, a_desc0);
	TDL_remap_ref(b_desc, b_desc0);
	TDL_remap_ref(c_desc, c_desc0);
	TDL_fence();

#ifdef TIME_ONLY_COMPUTATION
{
    ttime_t t1, t2;
    tgettime(t1);
#endif

	for (i = l_start; i < l_end; i += l_step) {
	    double t = 0;
	    for (j = ib[i]; j < ib[i + 1]; ++j)
		t += a[j] * b[idx[j]];
	    c[i] += t;
	}

#ifdef TIME_ONLY_COMPUTATION
    tgettime(t2);
    *time = ttime_diff_sec(t1, t2);
}
#endif

	/**  Update  **/

	TDL_fence();
	TDL_update_modify(c_desc);
	TDL_fence();
    }
}
