#include <math.h>
#include <stdio.h>

/*
 *  2-dimensional Poisson Equation with Dirichlet Condition
 *
 *  
 */

void
prob1(n, m, ai, aj, a, size, nzero)
    int n, m;
    int *ai, *aj;
    double *a;
    int *size, *nzero;
{
    int i, j;
    double inv_h = (m + 1) * (m + 1);
    double d0 = 4.0 * inv_h;
    double d1 = -1.0 * inv_h;
    int N = n * m;
    int k = 0;

    /**  Diagonal elements  **/

    for (i = 0; i < N; ++i) {
/*	int k = i; */
	a[k] = d0;
	ai[k] = aj[k] = i;
	++k;
    }

    /**  Sub-diagonal elements  **/

    for (i = 0; i < N; i += m) {
	for (j = i; j < i + m - 1; ++j) {
/*	    int k = N + 2 * ((i / m) * (m - 1) + (j - i));  */
	    a[k] = d1;
	    ai[k] = j + 1;
	    aj[k] = ai[k] - 1;
	    ++k;
	    a[k] = d1;
	    ai[k] = j;
	    aj[k] = ai[k] + 1;
	    ++k;
	}
    }

    for (i = m; i < N; ++i) {
/*	int k = N + 2 * (m - 1) * n + 2 * (i - m); */
	a[k] = d1;
	ai[k] = i;
	aj[k] = ai[k] - m;
	++k;
	a[k] = d1;
	ai[k] = i - m;
	aj[k] = ai[k] + m;
	++k;
    }

    *size = N;
    *nzero = k;
}

/*
 *  2-dimensional Poisson Equation with Dirichlet Condition
 *
 *  __________
 *  |   m1   |
 *  |        |
 *  |  m2  |    n
 *  |      |
 *    ~~~~~~
 */

void
prob2(n, m1, m2, ai, aj, a, size, nzero)
    int n, m1, m2;
    int *ai, *aj;
    double *a;
    int *size, *nzero;
{
    int i, j, k;
    double inv_h = (n + 1) * (m1 + 1);
    double d0 = 4.0 * inv_h;
    double d1 = -1.0 * inv_h;
    int n2 = n / 2;
    int N2 = m1 * n2;
    int N = N2 + m2 * (n - n2);
    int min_m = m1 < m2 ? m1 : m2;

    k = 0;

    /**  Diagonal elements  **/

    for (i = 0; i < N; ++i) {
	a[k] = d0;
	ai[k] = aj[k] = i;
	++k;
    }

    /**  Sub-diagonal elements  **/

    for (i = 0; i < n2; ++i) {
	for (j = 0; j < m1 - 1; ++j) {
	    int l = i * m1 + j;
	    a[k] = d1;
	    ai[k] = l + 1;
	    aj[k] = ai[k] - 1;
	    ++k;
	    a[k] = d1;
	    ai[k] = l;
	    aj[k] = ai[k] + 1;
	    ++k;
	}
    }
    for (i = 0; i < n - n2; ++i) {
	for (j = 0; j < m2 - 1; ++j) {
	    int l = N2 + i * m2 + j;
	    a[k] = d1;
	    ai[k] = l + 1;
	    aj[k] = ai[k] - 1;
	    ++k;
	    a[k] = d1;
	    ai[k] = l;
	    aj[k] = ai[k] + 1;
	    ++k;
	}
    }

    for (i = 0; i < N2 - m1 + min_m; ++i) {
	a[k] = d1;
	ai[k] = i + m1;
	aj[k] = ai[k] - m1;
	++k;
	a[k] = d1;
	ai[k] = i;
	aj[k] = ai[k] + m1;
	++k;
    }
    for (i = N2; i < N - m2; ++i) {
	a[k] = d1;
	ai[k] = i + m2;
	aj[k] = ai[k] - m2;
	++k;
	a[k] = d1;
	ai[k] = i;
	aj[k] = ai[k] + m2;
	++k;
    }

    *size = N;
    *nzero = k;
}


/*
 *  2-dimensional Poisson Equation with Dirichlet Condition
 *
 *  Spiral Ordering
 */

void
r_to_ij(n, r, i, j)
    int n, r;
    int *i, *j;
{
    int c, d, e, f;
    double d_n = (double)n;

    c = (int)floor(.5 * (d_n - sqrt(d_n * d_n - r)));
    d = r - 4 * c * (n - c);
    e = n - 2 * c - 1;
    if (e > 0)
	f = d / e;
    else
	f = 0;

    switch (f) {
    case 0:  /**  upper edge  **/
	*i = c;
	*j = c + d;
	break;
    case 1:  /**  right edge  **/
	*i = c + d - e;
	*j = n - 1 - c;
	break;
    case 2:  /**  lower edge  **/
	*i = n - 1 - c;
	*j = n - 1 - c - (d - 2 * e);
	break;
    case 3:  /**  left edge  **/
	*i = n - 1 - c - (d - 3 * e);
	*j = c;
	break;
    default:
	perror("");
	exit(1);
    }
}

#define MAX(i, j)  ((i) > (j) ? (i) : (j))

int
ij_to_r(n, i, j)
    int n, i, j;
{
    int i1, j1, c;
    int n2 = n / 2;

    i1 = i - n2;
    j1 = j - n2;
    c = n2 - MAX(abs(i1), abs(j1));

    if (i - j <= 0) {
	return 4 * c * (n - c) + i + j - 2 * c;
    }
    else {
	return 4 * (c + 1)* (n - c - 1) - (i + j - 2 * c);
    }
}


/*
 *  Spiral problem
 *
 *  Assume n == m
 */

void
prob3(n, m, ai, aj, a, size, nzero)
    int n, m;
    int *ai, *aj;
    double *a;
    int *size, *nzero;
{
    int i, k;
    double inv_h = (m + 1) * (m + 1);
    double d0 = 4.0 * inv_h;
    double d1 = -1.0 * inv_h;
    int N = n * m;

    k = 0;

    for (i = 0; i < N; ++i) {
	int s, t;

	/**  Diagonal elements  **/

	a[k] = d0;
	ai[k] = aj[k] = i;
	++k;

	/**  Subdiagonal elements  **/

	r_to_ij(n, i, &s, &t);
	
	if (s - 1 >= 0) {
	    a[k] = d1;
	    ai[k] = i;
	    aj[k] = ij_to_r(n, s - 1, t);
	    ++k;
	}
	if (t - 1 >= 0) {
	    a[k] = d1;
	    ai[k] = i;
	    aj[k] = ij_to_r(n, s, t - 1);
	    ++k;
	}
	if (s + 1 < n) {
	    a[k] = d1;
	    ai[k] = i;
	    aj[k] = ij_to_r(n, s + 1, t);
	    ++k;
	}
	if (t + 1 < n) {
	    a[k] = d1;
	    ai[k] = i;
	    aj[k] = ij_to_r(n, s, t + 1);
	    ++k;
	}
    }

    *size = N;
    *nzero = k;
}


#if 0

#define n	5
#define m	5
#define N	(n * m)
#define N5	(5 * N)

/*#include "mult.h"*/

main()
{
    int ai[N5], aj[N5];
    double a[N5];
    int idx[N5], ib[N], ie[N];
    double b[N5];
    int k;
    double c[N], d[N];
    double e, f;
    int mat_size;

    prob1(n, m, ai, aj, a, &mat_size, &k);
    print_COO(k, ai, aj, a);
#if 0
    prob2(n, m, m - 2, ai, aj, a, &mat_size, &k);
    print_COO(k, ai, aj, a);
#endif
#if 0
    prob3(n, m, ai, aj, a, &mat_size, &k);
    print_COO(k, ai, aj, a);
#endif
#if 0
    COO_to_CSR(mat_size, k, ai, aj, a, idx, ib, ie, b);

/*    print_CSR(mat_size, idx, ib, ie, b); */

/*    vec_mult_COO(mat_size, k, ai, aj, a, a, c); */
/*    vec_mult_CSR(mat_size, idx, ib, ie, b, a, d); */

/*    e = res_COO(mat_size, k, ai, aj, a, a, b, c); */
/*    f = res_CSR(mat_size, idx, ib, ie, b, a, b, d); */

    e = vec_mult_ip_COO(mat_size, k, ai, aj, a, a, c);
    f = vec_mult_ip_CSR(mat_size, idx, ib, ie, b, a, d);

    print_vec(mat_size, c);
    printf("(COO) : %g\n", e);
    print_vec(mat_size, d);
    printf("(CSR) : %g\n", f);
#endif
}

#endif
