/*
 *  $Id: tdl.c,v 1.1.1.1 2004-02-06 18:14:59 msato Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "tdl.h"
#include "tdl_allgather.h"
#include "tdl_debug.h"

/*
 *  Global variables
 */

MPI_Comm TDL_gComm;

int TDL_gNumThreads;
int TDL_gThreadNum;

/*
 *  Constants
 */

#ifdef REDIRECT_STDOUT
#define TDL_STDOUT_SUF	"-out."
#define TDL_MAXINTLEN	10
#endif /* REDIRECT_STDOUT */

/*
 *  External Functions
 */

void
isort(int n, Send_el a);

/*
 *  Static Function Declarations
 */

static TDL_desc
TDL_desc_create_size(int size, void *base);

static void
_TDL_desc_copy_sendrecv(TDL_desc src, TDL_desc dest);

static void
_TDL_desc_copy_dist(TDL_desc src, TDL_desc dest);

static void
_TDL_desc_free_mlist(TDL_desc di, TDL_Messages mlist);

static void
_TDL_desc_free_each_send_list(TDL_desc di);

static void
_TDL_desc_free_each_recv_list(TDL_desc di);

static void
_TDL_desc_free_home(TDL_desc di);

static void
_TDL_desc_free_sendrecv(TDL_desc di);

static void
_TDL_desc_free(TDL_desc desc);

static void
TDL_dist_master(TDL_desc tdl_desc);

static void
TDL_dist_master_all(TDL_desc tdl_desc);

static void
TDL_dist_block(TDL_desc tdl_desc);

static void
TDL_dist_block_all(TDL_desc tdl_desc);

static void
TDL_dist_cyclic(TDL_desc tdl_desc);

static int
TDL_sharing_num(TDL_desc tdl_desc, int page);

static int*
TDL_sharing_rank(TDL_desc tdl_desc, int page);

static void
TDL_add_list(Send_el_list *list, void *addr, int size,
	     TDL_desc tdl_desc);

#if 0
static int
TDL_conv_list_array(Send_el_list sl, Send_el *sll);

static Send_el_list
TDL_conv_array_list(int n, Send_el sll);

static int
TDL_compact_send_array(int n, Send_el sll);
#endif

#if 0
static void
TDL_message_aggregation(TDL_desc tdl_desc, Send_el_list *list);
#endif

static int
TDL_conv_msgs_array(TDL_desc desc, int page, Send_el *sll);

static int
TDL_check_intersection(int n, Send_el sll);

static void
TDL_check_recv_list(TDL_desc tdl_desc);

static void
TDL_create_recv_list(TDL_desc tdl_desc,
		     Send_el_list *send_list,
		     TDL_Messages recv_list);

static void
TDL_decide_home(TDL_desc tdl_desc);

static void
TDL_decide_sharing(TDL_desc tdl_desc);

static void
TDL_update_ref(TDL_desc tdl_desc);

static int
_TDL_remap_home(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc,
		int force);

/*
 *   Initialization/Finalization
 */

void
_TDL_init(char *progname)
{
    MPI_Comm_dup(MPI_COMM_WORLD, &TDL_gComm);

    MPI_Comm_size(TDL_gComm, &TDL_gNumThreads);
    MPI_Comm_rank(TDL_gComm, &TDL_gThreadNum);

    /**  Assertions  **/

    assert(TDL_gNumThreads <= sizeof(TDL_Share_type) * 8);

    /**  TDL_PAGE_MASK  **/

    assert(sizeof(unsigned long) == sizeof(void*));

    /**  In TDL_decide_sharing(tdl_desc)  **/

    {
	int size;
	MPI_Type_size(MPI_UNSIGNED_LONG, &size);
	assert(sizeof(TDL_Share_type) == size);
    }

#ifdef REDIRECT_STDOUT
    /**  Redirect stdout  **/

    {
	char *s, *out;
	int fd;

	s = progname;
	out = (char*)malloc(strlen(s) + strlen(TDL_STDOUT_SUF) +
			    TDL_MAXINTLEN);
	if (out == NULL)
	    fprintf(stderr, "TDL_init: not enough memory\n"), exit(1);
	sprintf(out, "%s%s%d", s, TDL_STDOUT_SUF, TDL_gThreadNum);

	if ((fd = creat(out, 0644)) == -1)
	    perror(out);
	else if (dup2(fd, 1) == -1)
	    perror("dup2");
    }
#endif /* REDIRECT_STDOUT */
}

void
TDL_init(int *argc, char **argv[])
{
    MPI_Init(argc, argv);

    _TDL_init(**argv);
}

void
TDL_finalize()
{
    MPI_Comm_free(&TDL_gComm);
    MPI_Finalize();
}

/*
 *  Each distributed array needs two arrays for descriptor:
 *
 *    TDL_ACCESS_COUNT	unsigned long[NUM_PAGES]  --- access counter & home
 *    TDL_SHARING	TDL_Share_type[NUM_PAGES] --- sharing information
 *
 *  For MPI-2 & MPI-1
 *
 *    TDL_SEND_LIST	Send_el_list[NUM_PAGES] --- list of updated elems
 *    TDL_RECV_LIST	Send_el_list[NUM_PAGES] --- list of referenced elems
 *
 *  For MPI-1
 *
 *    TDL_SEND_MLIST	TDL_Messages[NUM_PAGES]
 *    TDL_RECV_MLIST	TDL_Messages[NUM_PAGES]
 *			--- list of #msgs, addr, size for all processes
 */

static TDL_desc
TDL_desc_create_size(int size, void *base)
{
    TDL_desc di;
    int num_pages;

    di = (TDL_desc)malloc(sizeof(struct _TDL_desc));

    TDL_SIZE(di) = size;
    TDL_NUM_PAGES(di) = num_pages = TDL_NUM_PAGES_FROM_SIZE(size);
    TDL_BASE(di) = base;

    TDL_ACCESS_BASE(di) = (TDL_Access_count)calloc(1, sizeof(struct _TDL_access_count));
    TDL_ACCESS_COUNT(di) = (unsigned long *)calloc(num_pages, sizeof(unsigned long));
    TDL_PREV_BASE(di) = (TDL_Access_count)calloc(1, sizeof(struct _TDL_access_count));
    TDL_PREV_HOME(di) = (unsigned long *)malloc(num_pages * sizeof(unsigned long));
    {
	int i;
	for (i = 0; i < num_pages; ++i)
	    TDL_PREV_HOME(di)[i] = TDL_PAGE_UNDEF;
    }
    TDL_SHARE_BASE(di) = (TDL_Share)calloc(1, sizeof(struct _TDL_share));
    TDL_SHARING(di) = (TDL_Share_type *)calloc(num_pages, sizeof(TDL_Share_type));

    /**  TDL_SEND_LIST is necessary for MPI and MPI-2.  **/

    TDL_SEND_BASE(di) = (TDL_Msg_list)calloc(1, sizeof(struct _TDL_msg_list));
    TDL_SEND_LIST(di) = (Send_el_list *)calloc(num_pages, sizeof(Send_el_list));
#ifdef MPI1
    TDL_SEND_MLIST(di) = (TDL_Messages)calloc(num_pages, sizeof(struct _TDL_messages));
#endif

    TDL_RECV_BASE(di) = (TDL_Msg_list)calloc(1, sizeof(struct _TDL_msg_list));
    TDL_RECV_LIST(di) = (Send_el_list *)calloc(num_pages, sizeof(Send_el_list));
#ifdef MPI1
    TDL_RECV_MLIST(di) = (TDL_Messages)calloc(num_pages, sizeof(struct _TDL_messages));
#endif

    return di;
}

TDL_desc
TDL_desc_create_desc(TDL_desc tdl_desc)
{
    TDL_desc new_desc =
	TDL_desc_create_size(TDL_SIZE(tdl_desc), TDL_BASE(tdl_desc));

    return new_desc;
}

static void
_TDL_desc_copy_sendrecv(TDL_desc src, TDL_desc dest)
{
    _TDL_desc_free_sendrecv(dest);

    ++TDL_REF(TDL_SEND_BASE(src));
    TDL_SEND_BASE(dest) = TDL_SEND_BASE(src);

    ++TDL_REF(TDL_RECV_BASE(src));
    TDL_RECV_BASE(dest) = TDL_RECV_BASE(src);
}

static void
_TDL_desc_copy_dist(TDL_desc src, TDL_desc dest)
{
    assert(TDL_NUM_PAGES(src) == TDL_NUM_PAGES(dest));

    /* BASE address and GID are not copied. */

    _TDL_desc_free_home(dest);

    ++TDL_REF(TDL_ACCESS_BASE(src));
    TDL_ACCESS_BASE(dest) = TDL_ACCESS_BASE(src);
    
    ++TDL_REF(TDL_SHARE_BASE(src));
    TDL_SHARE_BASE(dest) = TDL_SHARE_BASE(src);

    _TDL_desc_copy_sendrecv(src, dest);    
}

static void
_TDL_desc_free_mlist(TDL_desc di, TDL_Messages mlist)
{
    int i;
    for (i = 0; i < TDL_NUM_PAGES(di); ++i) {
	if (mlist[i].num != NULL) {
	    free(mlist[i].num);
	    free(mlist[i].offset);
	    free(mlist[i].size);
	}
    }
}

static void
_TDL_desc_free_send_list(TDL_desc tdl_desc,
			 Send_el_list *send_list,
			 TDL_Messages recv_list)
{
    int i;
    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	delete_send_el(send_list[i]);
    }
#ifdef MPI1
    _TDL_desc_free_mlist(tdl_desc, recv_list);
#endif
}

static void
_TDL_desc_free_each_send_list(TDL_desc di)
{
    _TDL_desc_free_send_list(di, TDL_SEND_LIST(di),
			     TDL_SEND_MLIST(di));
}

static void
_TDL_desc_free_each_recv_list(TDL_desc di)
{
    _TDL_desc_free_send_list(di, TDL_RECV_LIST(di),
			     TDL_RECV_MLIST(di));
}

static void
_TDL_desc_free_home(TDL_desc di)
{
    if (TDL_REF(TDL_ACCESS_BASE(di))-- == 0) {
	free(TDL_ACCESS_COUNT(di));
	free(TDL_ACCESS_BASE(di));
    }
    if (TDL_REF(TDL_PREV_BASE(di))-- == 0) {
	free(TDL_PREV_HOME(di));
	free(TDL_PREV_BASE(di));
    }
    if (TDL_REF(TDL_SHARE_BASE(di))-- == 0) {
	free(TDL_SHARING(di));
	free(TDL_SHARE_BASE(di));
    }
}

static void
_TDL_desc_free_sendrecv(TDL_desc di)
{
    if (TDL_REF(TDL_SEND_BASE(di))-- == 0) {

	_TDL_desc_free_each_send_list(di);

	free(TDL_SEND_LIST(di));
#ifdef MPI1
	free(TDL_SEND_MLIST(di));
#endif
	free(TDL_SEND_BASE(di));
    }

    if (TDL_REF(TDL_RECV_BASE(di))-- == 0) {

	_TDL_desc_free_each_recv_list(di);

	free(TDL_RECV_LIST(di));
#ifdef MPI1
	free(TDL_RECV_MLIST(di));
#endif
	free(TDL_RECV_BASE(di));
    }
}

static void
_TDL_desc_free(TDL_desc di)
{
    _TDL_desc_free_home(di);
    _TDL_desc_free_sendrecv(di);
}

void
TDL_desc_free(TDL_desc di)
{
    _TDL_desc_free(di);
    free(di);
}

/*
 *  Print TDL Information
 */

void
TDL_desc_print(TDL_desc tdl_desc)
{
    int i;

    printf("SIZE: %d, PNUM: %d, BASE %p\n",
	   TDL_SIZE(tdl_desc), TDL_NUM_PAGES(tdl_desc), TDL_BASE(tdl_desc));
    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	printf("[%d] PAGE %d: HOME %ld, SHARE %lo", TDL_gThreadNum, i,
	       TDL_HOME(tdl_desc)[i], TDL_SHARING(tdl_desc)[i]);
	if (TDL_sharing_num(tdl_desc, i) > 1) {
	    int *sharing_rank = TDL_sharing_rank(tdl_desc, i);
	    int j;
	    printf(" (%d", sharing_rank[0]);
	    for (j = 1; j < TDL_sharing_num(tdl_desc, i); ++j) {
		printf(" %d", sharing_rank[j]);
	    }
	    printf(")");
	    free(sharing_rank);
	}
	printf("\n");

	print_send_el(TDL_SEND_LIST(tdl_desc)[i]);
#ifdef MPI1
	if (TDL_IS_SHARED(tdl_desc, i) && TDL_SEND_MLIST(tdl_desc)[i].num) {
	    int j, jj = 0;
	    for (j = 0; j < TDL_sharing_num(tdl_desc, i); ++j) {
		int k, *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		for (k = 0; k < TDL_SEND_MLIST(tdl_desc)[i].num[j]; ++k) {
		    printf("[%d] PAGE %d: RECV (RANK %d; ADDR %lx; SIZE %lu)\n",
			   TDL_gThreadNum, i, sharing_rank[j],
			   TDL_SEND_MLIST(tdl_desc)[i].offset[jj],
			   TDL_SEND_MLIST(tdl_desc)[i].size[jj]);
		    jj++;
		}
		free(sharing_rank);
	    }
	}
#endif

	print_send_el(TDL_RECV_LIST(tdl_desc)[i]);
#ifdef MPI1
	if (TDL_IS_SHARED(tdl_desc, i) && TDL_RECV_MLIST(tdl_desc)[i].num) {
	    int j, jj = 0;
	    for (j = 0; j < TDL_sharing_num(tdl_desc, i); ++j) {
		int k, *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		for (k = 0; k < TDL_RECV_MLIST(tdl_desc)[i].num[j]; ++k) {
		    printf("[%d] PAGE %d: SEND (RANK %d; ADDR %lx; SIZE %lu)\n",
			   TDL_gThreadNum, i, sharing_rank[j],
			   TDL_RECV_MLIST(tdl_desc)[i].offset[jj],
			   TDL_RECV_MLIST(tdl_desc)[i].size[jj]);
		    jj++;
		}
		free(sharing_rank);
	    }
	}
#endif
    }
}

void
TDL_desc_print_abs(TDL_desc tdl_desc)
{
    int i;
    unsigned long prev_home = TDL_PAGE_UNDEF;
    TDL_Share_type prev_share = TDL_SHARE_UNDEF;
    int first_time = 1;

    printf("SIZE: %d, PNUM: %d, BASE %p\n",
	   TDL_SIZE(tdl_desc), TDL_NUM_PAGES(tdl_desc), TDL_BASE(tdl_desc));
    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {

	if ((TDL_HOME(tdl_desc)[i] != prev_home
	     || TDL_SHARING(tdl_desc)[i] != prev_share)  /* first case */
	    || i == TDL_NUM_PAGES(tdl_desc) - 1          /* last */
	    || (TDL_HOME(tdl_desc)[i + 1] != prev_home   /* last */
		|| TDL_SHARING(tdl_desc)[i + 1] != prev_share)) {

	    first_time = 1;
	    printf("[%d] PAGE %d: HOME %ld, SHARE %lo", TDL_gThreadNum, i,
		   TDL_HOME(tdl_desc)[i], TDL_SHARING(tdl_desc)[i]);
	    if (TDL_sharing_num(tdl_desc, i) > 1) {
		int *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		int j;
		printf(" (%d", sharing_rank[0]);
		for (j = 1; j < TDL_sharing_num(tdl_desc, i); ++j) {
		    printf(" %d", sharing_rank[j]);
		}
		printf(")");
		free(sharing_rank);
	    }
	    printf("\n");

	    prev_home = TDL_HOME(tdl_desc)[i];
	    prev_share = TDL_SHARING(tdl_desc)[i];
	}
	else {
	    if (first_time) {
		printf("[%d] ...\n", TDL_gThreadNum);
		first_time = 0;
	    }
	}

	print_send_el(TDL_SEND_LIST(tdl_desc)[i]);
#ifdef MPI1
	if (TDL_IS_SHARED(tdl_desc, i) && TDL_SEND_MLIST(tdl_desc)[i].num) {
	    int j, jj = 0;
	    for (j = 0; j < TDL_sharing_num(tdl_desc, i); ++j) {
		int k, *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		for (k = 0; k < TDL_SEND_MLIST(tdl_desc)[i].num[j]; ++k) {
		    printf("[%d] PAGE %d: RECV (RANK %d; ADDR %lx; SIZE %lu)\n",
			   TDL_gThreadNum, i, sharing_rank[j],
			   TDL_SEND_MLIST(tdl_desc)[i].offset[jj],
			   TDL_SEND_MLIST(tdl_desc)[i].size[jj]);
		    jj++;
		}
		free(sharing_rank);
	    }
	}
#endif

	print_send_el(TDL_RECV_LIST(tdl_desc)[i]);
#ifdef MPI1
	if (TDL_IS_SHARED(tdl_desc, i) && TDL_RECV_MLIST(tdl_desc)[i].num) {
	    int j, jj = 0;
	    for (j = 0; j < TDL_sharing_num(tdl_desc, i); ++j) {
		int k, *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		for (k = 0; k < TDL_RECV_MLIST(tdl_desc)[i].num[j]; ++k) {
		    printf("[%d] PAGE %d: SEND (RANK %d; ADDR %lx; SIZE %lu)\n",
			   TDL_gThreadNum, i, sharing_rank[j],
			   TDL_RECV_MLIST(tdl_desc)[i].offset[jj],
			   TDL_RECV_MLIST(tdl_desc)[i].size[jj]);
		    jj++;
		}
		free(sharing_rank);
	    }
	}
#endif
    }
}

#if 0 /* Not implemented yet */
void
TDL_print_array_double(TDL_desc tdl_desc)
{
    int i, j;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_HOME_PAGE(tdl_desc, i)) {
	    int start = TDL_PAGE_OFFSET(i);
	    int offset = TDL_PAGE_OFFSET(i + 1);
	    int end = (offset < TDL_SIZE(tdl_desc)) ? offset : TDL_SIZE(tdl_desc);

	    double *d = (double *)TDL_ADDR(start, tdl_desc);
	    start /= sizeof(double);
	    end /= sizeof(double);
	}
}
#endif

/*
 *  MASTER/ALL/BLOCK/CYCLIC Distribution
 */

static void
TDL_dist_master(TDL_desc tdl_desc)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	TDL_HOME(tdl_desc)[i] = 0;
	TDL_SHARING(tdl_desc)[i] = 1 << 0;
    }
}

static void
TDL_dist_master_all(TDL_desc tdl_desc)
{
    int i;
    TDL_Share_type sh = 0;

    for (i = 0; i < TDL_gNumThreads; ++i)
	sh |= 1 << i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	TDL_HOME(tdl_desc)[i] = 0;
	TDL_SHARING(tdl_desc)[i] = sh;
    }
}

static void
TDL_dist_block(TDL_desc tdl_desc)
{
    int i, j;
    int psize = TDL_NUM_PAGES(tdl_desc);

    for (i = 0; i < TDL_gNumThreads; ++i) {
	for (j = psize * i / TDL_gNumThreads;
	     j < psize * (i + 1) / TDL_gNumThreads;
	     ++j) {
	    TDL_HOME(tdl_desc)[j] = i;
	    TDL_SHARING(tdl_desc)[j] = 1 << i;
	}
    }
}

static void
TDL_dist_block_all(TDL_desc tdl_desc)
{
    int i, j;
    TDL_Share_type sh = 0;
    int psize = TDL_NUM_PAGES(tdl_desc);

    for (i = 0; i < TDL_gNumThreads; ++i)
	sh |= 1 << i;

    for (i = 0; i < TDL_gNumThreads; ++i) {
	for (j = psize * i / TDL_gNumThreads;
	     j < psize * (i + 1) / TDL_gNumThreads;
	     ++j) {
	    TDL_HOME(tdl_desc)[j] = i;
	    TDL_SHARING(tdl_desc)[j] = sh;
	}
    }
}

static void
TDL_dist_cyclic(TDL_desc tdl_desc)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	TDL_HOME(tdl_desc)[i] = i % TDL_gNumThreads;
	TDL_SHARING(tdl_desc)[i] = 1 << (i % TDL_gNumThreads);
    }
}

void
TDL_desc_dist(TDL_desc tdl_desc, TDL_dist dist)
{
    _TDL_desc_free_each_send_list(tdl_desc);
    _TDL_desc_free_each_recv_list(tdl_desc);

    switch(dist){
    case TDL_MASTER:
	TDL_dist_master(tdl_desc);
	break;
    case TDL_MASTER_ALL:
	TDL_dist_master_all(tdl_desc);
	break;
    case TDL_BLOCK:
	TDL_dist_block(tdl_desc);
	break;
    case TDL_BLOCK_ALL:
	TDL_dist_block_all(tdl_desc);
	break;
    case TDL_CYCLIC:
	TDL_dist_cyclic(tdl_desc);
	break;
    default:
	fprintf(stderr, "TDL_desc_dist(): Not supported yet\n");
	fflush(stderr);
	exit(1);
    }
}

void
TDL_desc_copy_dist(TDL_desc src, TDL_desc dest)
{
    int i;

    assert(TDL_NUM_PAGES(src) == TDL_NUM_PAGES(dest));

    for (i = 0; i < TDL_NUM_PAGES(dest); ++i) {
	TDL_HOME(dest)[i] = TDL_HOME(src)[i];
	TDL_SHARING(dest)[i] = TDL_SHARING(src)[i];
    }
}

void
TDL_desc_copy_home(TDL_desc src, TDL_desc dest)
{
    int i;

    assert(TDL_NUM_PAGES(src) == TDL_NUM_PAGES(dest));

    for (i = 0; i < TDL_NUM_PAGES(dest); ++i) {
	TDL_HOME(dest)[i] = TDL_HOME(src)[i];
	TDL_SHARING(dest)[i] |= 1 << TDL_HOME(src)[i];
    }
}

int
TDL_desc_compare_dist(TDL_desc src, TDL_desc dest)
{
    int i;

    assert(TDL_NUM_PAGES(src) == TDL_NUM_PAGES(dest));

    for (i = 0; i < TDL_NUM_PAGES(dest); ++i) {
	if (TDL_HOME(dest)[i] != TDL_HOME(src)[i]
	    || TDL_SHARING(dest)[i] != TDL_SHARING(src)[i])
	    return 0;
    }
    return 1;
}

int
TDL_desc_compare_home(TDL_desc src, TDL_desc dest)
{
    int i;

    assert(TDL_NUM_PAGES(src) == TDL_NUM_PAGES(dest));

    for (i = 0; i < TDL_NUM_PAGES(dest); ++i) {
	if (TDL_HOME(dest)[i] != TDL_HOME(src)[i])
	    return 0;
    }
    return 1;
}

/*
 *  TDL_malloc():
 *
 *  This function is obsolete.  Now, it is not necessary that each
 *  distributed array is page-aligned.
 *
 *  This function allocates memory and returns page-aligned address.
 */

void*
TDL_malloc(int byte_size)
{
    void *base;

    fprintf(stderr, "TDL_malloc() is obsolete.  Use malloc() instead\n");
    fflush(stderr);

    /*****
    *****   Now, 2 * TDL_PAGE_SIZE is allocated.
    *****   Digital UNIX causes Segmentation Fault in redistribute
    *****   since redistribute always send/recv PAGE size.
    ******/

    /*****
    *****   Linux requires the allocated address to free the allocated 
    *****   memory.  However, it is not considered yet.
    ******/

    base = malloc(byte_size + 2 * TDL_PAGE_SIZE - 1);
    return TDL_PAGE_ALIGN(base);
}

/*
 *  TDL_desc_create()
 */

TDL_desc
TDL_desc_create(void *base, int byte_size, TDL_dist dist)
{
    TDL_desc t_desc;

    /**  Create TDL Desc. **/

    t_desc = TDL_desc_create_size(byte_size, base);

    TDL_desc_dist(t_desc, dist);

    /*
     *  It is necessary to create MPI_Win for MPI-2 one-sided
     *  communication.
     */

    /** **/

    debug(printf("[%d] TDL_desc_create: base = %p, size = %d\n",
		 TDL_gThreadNum, base, byte_size));
    debug(fflush(stdout));

    return t_desc;
}

/*
 *  Misc.
 */

static int
TDL_sharing_num(TDL_desc tdl_desc, int page)
{
    int i, k;

    k = 0;
    for (i = 0; i < TDL_gNumThreads; ++i) {
	if (TDL_IS_SHARED_WITH(tdl_desc, page, i))
	    ++k;
    }
    return k;
}

static int*
TDL_sharing_rank(TDL_desc tdl_desc, int page)
{
    int i, k, n;
    int *t;

    n = TDL_sharing_num(tdl_desc, page);
    t = (int*)malloc(sizeof(int) * n);

    k = 0;
    for (i = 0; i < TDL_gNumThreads; ++i) {
	if (TDL_IS_SHARED_WITH(tdl_desc, page, i)) {
	    t[k] = i;
	    ++k;
	}
    }
    return t;    /*  free() is necessary.  */
}

/*
 *  TDL_inspect_home()
 */

void
TDL_inspect_home(void *addr, int size, TDL_desc desc)
{
    char *caddr = (char*)addr;
    int pnum = TDL_PAGE(caddr, desc);

    assert(TDL_DISP(caddr + size, desc) <= TDL_SIZE(desc));

    while (TDL_PAGE(caddr + size - 1, desc) != pnum) {
	int s = TDL_PAGE_OFFSET(pnum + 1) - TDL_DISP(caddr, desc);

	assert(pnum < TDL_NUM_PAGES(desc));

	TDL_ACCESS_COUNT(desc)[pnum] += s;
	TDL_SHARING(desc)[pnum] |= 1 << TDL_gThreadNum;

	caddr += s;
	size -= s;

	++pnum;
    }
    assert(pnum < TDL_NUM_PAGES(desc));

    TDL_ACCESS_COUNT(desc)[pnum] += size;
    TDL_SHARING(desc)[pnum] |= 1 << TDL_gThreadNum;
}

void
TDL_inspect_sharing(void *addr, int size, TDL_desc desc)
{
    char *caddr = (char*)addr;
    int pnum = TDL_PAGE(caddr, desc);

    assert(TDL_DISP(caddr + size, desc) <= TDL_SIZE(desc));

    while (TDL_PAGE(caddr + size - 1, desc) != pnum) {
	int s = TDL_PAGE_OFFSET(pnum + 1) - TDL_DISP(caddr, desc);

	assert(pnum < TDL_NUM_PAGES(desc));

	TDL_SHARING(desc)[pnum] |= 1 << TDL_gThreadNum;

	caddr += s;
	size -= s;

	++pnum;
    }
    assert(pnum < TDL_NUM_PAGES(desc));

    TDL_SHARING(desc)[pnum] |= 1 << TDL_gThreadNum;
}

static void
TDL_add_list(Send_el_list *list, void *addr, int size,
	     TDL_desc desc)
{
    char *caddr = (char*)addr;
    int pnum = TDL_PAGE(caddr, desc);
    struct _send_el sl;

    assert(TDL_DISP(caddr + size, desc) <= TDL_SIZE(desc));

    /**  Add to a send list  **/

    while (TDL_PAGE(caddr + size - 1, desc) != pnum) {
	sl.addr = TDL_DISP(caddr, desc);
	sl.size = TDL_PAGE_OFFSET(pnum + 1) - TDL_DISP(caddr, desc);

	assert(pnum < TDL_NUM_PAGES(desc));

/*	if (!(TDL_IS_PRIVATE(desc, pnum))) */
	    list[pnum] = insert_sorted_send_el(&sl, list[pnum]);

	caddr += sl.size;
	size -= sl.size;

	++pnum;
    }
    sl.addr = TDL_DISP(caddr, desc);
    sl.size = size;

    assert(pnum < TDL_NUM_PAGES(desc));

/*    if (!(TDL_IS_PRIVATE(desc, pnum))) */
	list[pnum] = insert_sorted_send_el(&sl, list[pnum]);
}

/*
 *  TDL_inspect_modify()
 */

void
TDL_inspect_modify(void *addr, int size, TDL_desc tdl_desc)
{
    TDL_add_list(TDL_SEND_LIST(tdl_desc), addr, size, tdl_desc);
}

/*
 *  TDL_inspect_ref()
 */

void
TDL_inspect_ref(void *addr, int size, TDL_desc tdl_desc)
{
    TDL_add_list(TDL_RECV_LIST(tdl_desc), addr, size, tdl_desc);
}

/*
 *  Convert list to array
 */

#if 0
static int
TDL_conv_list_array(Send_el_list sl, Send_el *sll)
{
    int n = length_send_el(sl);
    Send_el t_sll;

    *sll = t_sll = (Send_el)malloc(sizeof(struct _send_el) * n);
    while (sl) {
	*(t_sll++) = *(car_send_el(sl));
	sl = cdr_send_el(sl);
    }
    /*  sll needs to be freed later. */
    return n;
}

static Send_el_list
TDL_conv_array_list(int n, Send_el sll)
{
    if (n > 0) {
	return cons_send_el(sll, TDL_conv_array_list(n - 1, sll + 1));
    }
    else
	return NULL;
}

static int
TDL_compact_send_array(int n, Send_el sll)
{
    int i, j;
    Send_el p1, p2;

    j = 0;
    p1 = &sll[j];

    for (i = 1; i < n; ++i) {
	p2 = &sll[i];
	
	if (p1->addr + p1->size >= p2->addr) {
	    unsigned long new_size = p2->addr + p2->size - p1->addr;
	    
	    if (p1->size < new_size) {
		p1->size = new_size;
	    }
	}
	else {
	    ++j;
	    if (j < i)
		sll[j] = sll[i];
	    p1 = &sll[j];
	}
    }

    return (j + 1) < n ? (j + 1) : n;
}
#endif

/*
 *  Message Aggregation
 */

#if 0
static void
TDL_message_aggregation(TDL_desc tdl_desc, Send_el_list *list)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	int len;
	Send_el tmp_sll; /*  an array of struct _send_el_list  */

	/*  Convert a list to an array. */

	len = TDL_conv_list_array(list[i], &tmp_sll);
	delete_send_el(list[i]);

	/*  Sort and compact the array. */

	isort(len, tmp_sll);
	len = TDL_compact_send_array(len, tmp_sll);

	/*  Convert the array to a list. */

	list[i] = TDL_conv_array_list(len, tmp_sll);

	free(tmp_sll);
    }
}
#endif

/* */

static int
TDL_conv_msgs_array(TDL_desc desc, int page, Send_el *sll)
{
    int num_msgs, i;
    Send_el t_sll;

    num_msgs = 0;

    /*  This page should be a shared page. */

    assert(TDL_SEND_MLIST(desc)[page].num);

    for (i = 0; i < TDL_sharing_num(desc, page); ++i)
	num_msgs += TDL_SEND_MLIST(desc)[page].num[i];

    *sll = t_sll = (Send_el)malloc(sizeof(struct _send_el) * num_msgs);
    for (i = 0; i < num_msgs; ++i) {
	t_sll->addr = TDL_SEND_MLIST(desc)[page].offset[i];
	t_sll->size = TDL_SEND_MLIST(desc)[page].size[i];
	++t_sll;
    }
    /*  sll needs to be freed later. */
    return num_msgs;
}

static int
TDL_check_intersection(int n, Send_el sll)
{
    int i, j;
    Send_el p1, p2;

    j = 0;
    p1 = &sll[j];

    for (i = 1; i < n; ++i) {
	p2 = &sll[i];
	
	if (p1->addr + p1->size > p2->addr) {
	    return 1;
	}
	else {
	    ++j;
	    p1 = &sll[j];
	}
    }

    return 0;
}

/*
 *  Check a recv list whether the same address is not updated by more
 *  than two processes.
 */

static void
TDL_check_recv_list(TDL_desc tdl_desc)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {
	    int len;
	    Send_el tmp_sll; /*  an array of struct _send_el_list  */

	    /*  Convert TDL_Messages to an array of struct _send_el_list. */

	    len = TDL_conv_msgs_array(tdl_desc, i, &tmp_sll);

	    /*  Sort and check the intersection. */

	    isort(len, tmp_sll);

	    if (TDL_check_intersection(len, tmp_sll)) {
		fprintf(stderr, "[%d] Error: Store conflict!\n", TDL_gThreadNum);
		fflush(stderr);
		exit(1);
	    }

	    free(tmp_sll);
	}
    }
}

/*
 *  Create recv list
 */

static void
TDL_create_recv_list(TDL_desc tdl_desc,
		     Send_el_list *send_list,
		     TDL_Messages recv_list)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {
	    Send_el_list sl = send_list[i];
	    int n = length_send_el(sl);

	    unsigned long *send_addr = (unsigned long *)malloc(sizeof(unsigned long) * n);
	    unsigned long *send_size = (unsigned long *)malloc(sizeof(unsigned long) * n);
	    int *sharing_rank, total_num_recv;
	    int num_shared = TDL_sharing_num(tdl_desc, i);
	    int j;

	    sharing_rank = TDL_sharing_rank(tdl_desc, i);
	    recv_list[i].num = (int*)malloc(sizeof(int) * num_shared);

	    /**  Exchange size of send list  **/

	    debug(printf("[%d] TDL_allgather(%p (%d), 1, MPI_INT, %p, 1, MPI_INT, %d, %p, %d)\n",
			 TDL_gThreadNum, &n, n, recv_list[i].num,
			 num_shared, sharing_rank, TDL_CREATE_RECV1 + TDL_CREATE_SKIP * i));
	    debug(fflush(stdout));

	    TDL_allgather(&n, 1, MPI_INT,
			  recv_list[i].num, 1, MPI_INT,
			  num_shared, sharing_rank,
			  TDL_CREATE_RECV1 + TDL_CREATE_SKIP * i, TDL_gComm);

	    /**  Exchange send list  **/

	    j = 0;
	    while (sl) {
		Send_el p = car_send_el(sl);
		send_addr[j] = p->addr;
		send_size[j] = p->size;
		sl = cdr_send_el(sl);
		++j;
	    }
	    assert(j == n);

	    total_num_recv = 0;
	    for (j = 0; j < num_shared; ++j) {
		total_num_recv += recv_list[i].num[j];
	    }

	    recv_list[i].offset = (unsigned long *)malloc(sizeof(unsigned long) * total_num_recv);
	    recv_list[i].size = (unsigned long *)malloc(sizeof(unsigned long) * total_num_recv);

	    {
		int* displs = (int*)malloc(sizeof(int) * num_shared);
		int k;
		displs[0] = 0;
		for (k = 1; k < num_shared; ++k)
		    displs[k] = displs[k - 1] + recv_list[i].num[k - 1];

		debug(printf("[%d] TDL_allgatherv(%p, %d, MPI_UNSIGNED_LONG, %p, %p, %p, MPI_UNSIGNED_LONG, %d, %p, %d)\n",
			     TDL_gThreadNum, send_addr, n,
			     recv_list[i].offset, recv_list[i].num,
			     displs, num_shared, sharing_rank,
			     TDL_CREATE_RECV2 + TDL_CREATE_SKIP * i));
		debug(fflush(stdout));

		TDL_allgatherv(send_addr, n, MPI_UNSIGNED_LONG,
			       recv_list[i].offset, recv_list[i].num,
			       displs, MPI_UNSIGNED_LONG,
			       num_shared, sharing_rank,
			       TDL_CREATE_RECV2 + TDL_CREATE_SKIP * i,
			       TDL_gComm);

		debug(printf("[%d] TDL_allgatherv(%p, %d, MPI_UNSIGNED_LONG, %p, %p, %p, MPI_UNSIGNED_LONG, %d, %p, %d)\n",
			     TDL_gThreadNum, send_size, n,
			     recv_list[i].size, recv_list[i].num,
			     displs, num_shared, sharing_rank,
			     TDL_CREATE_RECV3 + TDL_CREATE_SKIP * i));
		debug(fflush(stdout));

		TDL_allgatherv(send_size, n, MPI_UNSIGNED_LONG,
			       recv_list[i].size, recv_list[i].num,
			       displs, MPI_UNSIGNED_LONG,
			       num_shared, sharing_rank,
			       TDL_CREATE_RECV3 + TDL_CREATE_SKIP * i, 
			       TDL_gComm);
	    
		debug(printf("[%d] TDL_allgatherv: done\n", TDL_gThreadNum));
		debug(fflush(stdout));

		free(displs);
	    }

	    free(sharing_rank);
	    free(send_addr);
	    free(send_size);
	}
/*	debug(printf("[%d] TDL_create_recv_list(): i = %d: done\n", TDL_gThreadNum, i)); */
    }

    debug(printf("[%d] TDL_create_recv_list(): done\n", TDL_gThreadNum));
    debug(fflush(stdout));
}

/*
 *  TDL_desc_fix_modify()
 */

void
TDL_desc_fix_modify(TDL_desc tdl_desc)
{
/*    TDL_message_aggregation(tdl_desc, TDL_SEND_LIST(tdl_desc)); */
#ifdef MPI1
    _TDL_desc_free_mlist(tdl_desc, TDL_SEND_MLIST(tdl_desc));
#endif
    TDL_create_recv_list(tdl_desc, TDL_SEND_LIST(tdl_desc),
			 TDL_SEND_MLIST(tdl_desc));
    TDL_check_recv_list(tdl_desc);
}

/*
 *  TDL_desc_fix_ref()
 */

void
TDL_desc_fix_ref(TDL_desc tdl_desc)
{
/*    TDL_message_aggregation(tdl_desc, TDL_RECV_LIST(tdl_desc)); */
#ifdef MPI1
    _TDL_desc_free_mlist(tdl_desc, TDL_RECV_MLIST(tdl_desc));
#endif
    TDL_create_recv_list(tdl_desc, TDL_RECV_LIST(tdl_desc),
			 TDL_RECV_MLIST(tdl_desc));
}

/*
 *  Decide sharing information
 */

struct int_int {
    int val;
    int rank;
};

static void
TDL_decide_home(TDL_desc tdl_desc)
{
    struct int_int *src, *recv;
    int psize = TDL_NUM_PAGES(tdl_desc);
    int i;

    src = (struct int_int *)malloc(sizeof(struct int_int) * psize);
    recv = (struct int_int *)malloc(sizeof(struct int_int) * psize);

    for (i = 0; i < psize; ++i) {
	src[i].val = (int)TDL_ACCESS_COUNT(tdl_desc)[i];
	src[i].rank = TDL_gThreadNum;
    }

    MPI_Allreduce(src, recv, psize, MPI_2INT, MPI_MAXLOC, TDL_gComm);

    for (i = 0; i < psize; ++i) {
	TDL_HOME(tdl_desc)[i] = recv[i].rank;
    }

    free(recv);
    free(src);
}

static void
TDL_decide_sharing(TDL_desc tdl_desc)
{
    int psize = TDL_NUM_PAGES(tdl_desc);
    TDL_Share_type *recv;
    int i;

    recv = (TDL_Share_type *)malloc(sizeof(TDL_Share_type) * psize);

    /**  MPI_UNSIGNED_LONG should be equivalent to TDL_Share_type. **/

    MPI_Allreduce(TDL_SHARING(tdl_desc), recv, psize,
		  MPI_UNSIGNED_LONG, MPI_BOR, TDL_gComm);

    for (i = 0; i < psize; ++i) {
	if (recv[i] == 0)
	    TDL_SHARING(tdl_desc)[i] = 1 << TDL_HOME(tdl_desc)[i];
	else
	    TDL_SHARING(tdl_desc)[i] = recv[i];
    }

    free(recv);
}

/*
 *  TDL_desc_fix_home()
 */

void
TDL_desc_fix_home(TDL_desc tdl_desc)
{
    TDL_decide_home(tdl_desc);
    TDL_decide_sharing(tdl_desc);
}

void
TDL_desc_fix_sharing(TDL_desc tdl_desc)
{
    TDL_decide_sharing(tdl_desc);
}

/*
 *  TDL_desc_merge()
 *
 *  If difference between tdl_desc1 and tdl_desc2 is small, the
 *  descriptor of tdl_desc1 is merged into tdl_desc2.
 */

/**  TDL_desc_merge() is not implemented enough.  **/

void
TDL_desc_merge(TDL_desc tdl_desc1, TDL_desc tdl_desc2)
{
    int i;

    fprintf(stderr, "TDL_desc_merge() is obsolete.\n");
    fflush(stderr);

    assert(TDL_NUM_PAGES(tdl_desc1) == TDL_NUM_PAGES(tdl_desc2));

    /*
     *  It is necessary to check difference of tdl_desc1 and
     *  tdl_desc2.  However, not implemented.
     */

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc1); ++i) {
	TDL_SHARING(tdl_desc2)[i] |= TDL_SHARING(tdl_desc1)[i];
    }

    /*
     *  If TDL_desc_merge() is used in an executor, copy or
     *  re-creation of recv/send list is necessary.
     */
}

/*
 *  Update/redistribute
 */

/*
 *  TDL_update():
 *
 *  Each modified data is multicasted to nodes that share the page.
 *
 *  Update schedule should be considered.
 *  Split-phase update is desirable.
 */

/*
 *  Vectorization (pack/unpack) is not implemented yet.
 */

void
TDL_update(TDL_desc tdl_desc)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {
	    int *sharing_rank = TDL_sharing_rank(tdl_desc, i);
	    int num_shared = TDL_sharing_num(tdl_desc, i);
	    int j, jj = 0;

	    assert(TDL_SEND_MLIST(tdl_desc)[i].num != NULL);

	    for (j = 0; j < num_shared; ++j) {
		/**  Update messages modified by each node. **/
		if (sharing_rank[j] == TDL_gThreadNum) {
		    int k;  /***  There are several messages to be sent.  ***/
		    for (k = 0; k < TDL_SEND_MLIST(tdl_desc)[i].num[j]; ++k) {
			int l;  /***  multicast  ***/
			for (l = 0; l < num_shared; ++l) {
			    if (sharing_rank[l] != TDL_gThreadNum) {

				debug(printf("[%d] MPI_Send(%p (%p + %lx), %lu, MPI_CHAR, %d, TDL_UPDATE)\n",
					     TDL_gThreadNum,
					     TDL_ADDR(TDL_SEND_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
					     TDL_BASE(tdl_desc), TDL_SEND_MLIST(tdl_desc)[i].offset[jj],
					     TDL_SEND_MLIST(tdl_desc)[i].size[jj],
					     sharing_rank[l]));
				debug(fflush(stdout));

				MPI_Send(TDL_ADDR(TDL_SEND_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
					 TDL_SEND_MLIST(tdl_desc)[i].size[jj], MPI_CHAR,
					 sharing_rank[l], TDL_UPDATE, TDL_gComm);
			    }
			}
			jj++;
		    }
		}
		else {
		    int k;
		    for (k = 0; k < TDL_SEND_MLIST(tdl_desc)[i].num[j]; ++k) {
			MPI_Status st;

			debug(printf("[%d] MPI_Recv(%p (%p + %lx), %lu, MPI_CHAR, %d, TDL_UPDATE)\n",
				     TDL_gThreadNum, TDL_ADDR(TDL_SEND_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
				     TDL_BASE(tdl_desc), TDL_SEND_MLIST(tdl_desc)[i].offset[jj],
				     TDL_SEND_MLIST(tdl_desc)[i].size[jj], sharing_rank[j]));
			debug(fflush(stdout));

			MPI_Recv(TDL_ADDR(TDL_SEND_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
				 TDL_SEND_MLIST(tdl_desc)[i].size[jj], MPI_CHAR,
				 sharing_rank[j], TDL_UPDATE,
				 TDL_gComm, &st);

			jj++;
		    }
		}
	    }
	    free(sharing_rank);
	}
/*	debug(printf("[%d] TDL_update: i = %d: done\n",	TDL_gThreadNum, i)); */
    }

#if 0 /* MPI2 or RHiNET */
    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {
	    int *sharing_rank = TDL_sharing_rank(tdl_desc, i);
	    int num_shared = TDL_sharing_num(tdl_desc, i);
	    Send_el_list sl = TDL_SEND_LIST(tdl_desc)[i];
#ifdef MPI2
#error Not implemented yet
	    while (sl) {
		Send_el p = car_send_el(sl);
		int j;
		/**  multicast  **/
		for (j = 0; j < num_shared; ++j)
		    if (sharing_rank[j] != TDL_gThreadNum)
			MPI_Put(TDL_ADDR(p->addr, tdl_desc), p->size, MPI_CHAR,
				sharing_rank[j], p->addr, p->size, MPI_CHAR,
				TDL_WIN(tdl_desc));

		sl = cdr_send_el(sl);
	    }
#elif defined(RHiNET)
#error Not implemented yet
	    if (TDL_SEND_LIST(tdl_desc)[i])
		mcast_diff();
#endif
	    free(sharing_rank);
	}
    }
#endif
    debug(printf("[%d] TDL_update: done\n", TDL_gThreadNum));
    debug(fflush(stdout));
}

/*
 *  TDL_update_ref():
 *
 *  Remap data that will be accessed only.
 */

static void
TDL_update_ref(TDL_desc tdl_desc)
{
    int i;
    int send_num = 0, send_byte = 0;
    int recv_num = 0, recv_byte = 0;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {

	/*  Necessary messages  */

	if (TDL_IS_SHARED(tdl_desc, i)) {

	    assert(TDL_RECV_MLIST(tdl_desc)[i].num != NULL);

	    if (TDL_IS_HOME_PAGE(tdl_desc, i)) {
		int *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		int num_shared = TDL_sharing_num(tdl_desc, i);
		int j, jj = 0;

		for (j = 0; j < num_shared; ++j) {
		    if (sharing_rank[j] != TDL_gThreadNum) {
			int k;  /***  There are several messages to be sent.  ***/
			for (k = 0; k < TDL_RECV_MLIST(tdl_desc)[i].num[j]; ++k) {
/*
			    debug(printf("[%d] MPI_Send(%p (%p + %lx), %lu, MPI_CHAR, %d, TDL_UPDATE_REF)\n",
					 TDL_gThreadNum,
					 TDL_ADDR(TDL_RECV_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
					 TDL_BASE(tdl_desc), TDL_RECV_MLIST(tdl_desc)[i].offset[jj],
					 TDL_RECV_MLIST(tdl_desc)[i].size[jj],
					 sharing_rank[j]));
*/
			    MPI_Send(TDL_ADDR(TDL_RECV_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
				     TDL_RECV_MLIST(tdl_desc)[i].size[jj], MPI_CHAR,
				     sharing_rank[j], TDL_UPDATE_REF, TDL_gComm);

			    debug(++send_num);
			    debug(send_byte += TDL_RECV_MLIST(tdl_desc)[i].size[jj]);

			    jj++;
			}
		    }
		    else {
			/*  It is not necessary to receive messages
			    since I am a home node.  */
			jj += TDL_RECV_MLIST(tdl_desc)[i].num[j];
		    }
		}
		free(sharing_rank);
	    }
	    else {
		Send_el_list rl = TDL_RECV_LIST(tdl_desc)[i];
		MPI_Status st;

		while (rl) {
		    Send_el p = car_send_el(rl);
/*
		    debug(printf("[%d] MPI_Recv(%p (%p + %lx), %lu, MPI_CHAR, %ld, TDL_UPDATE_REF)\n",
				 TDL_gThreadNum, TDL_ADDR(p->addr, tdl_desc),
				 TDL_BASE(tdl_desc), p->addr,
				 p->size, TDL_HOME(tdl_desc)[i]));
*/
		    MPI_Recv(TDL_ADDR(p->addr, tdl_desc),
			     p->size, MPI_CHAR,
			     TDL_HOME(tdl_desc)[i], TDL_UPDATE_REF,
			     TDL_gComm, &st);

		    debug(++recv_num);
		    debug(recv_byte += p->size);

		    rl = cdr_send_el(rl);
		}
	    }
	}
/*	debug(printf("[%d] TDL_update_ref: i = %d: done\n", TDL_gThreadNum, i)); */
    }

#if 0 /* MPI2 or RHiNET */
    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {
	    /*  If I'm not the home node. */
	    if (!(TDL_IS_HOME_PAGE(tdl_desc, i))) {
		Send_el_list rl = TDL_RECV_LIST(tdl_desc)[i];
		while (rl) {
		    Send_el p = car_send_el(rl);
#ifdef MPI2
#error Not implemented yet
		    MPI_Get(TDL_ADDR(p->addr, tdl_desc), p->size, MPI_CHAR,
			    TDL_HOME(tdl_desc)[i], p->addr, p->size, MPI_CHAR,
			    TDL_WIN(tdl_desc));
#elif defined(RHiNET)
#error Not implemented yet
#endif
		    rl = cdr_send_el(rl);
		}
	    }
	}
/*	debug(printf("[%d] TDL_update_ref: i = %d: done\n", TDL_gThreadNum, i)); */
    }
#endif
    if (send_num > 0 || recv_num > 0) {
	debug(printf("[%d] TDL_update_ref: base %p send %d byte (%d times), recv %d byte (%d times)\n",
		     TDL_gThreadNum, TDL_BASE(tdl_desc), send_byte, send_num, recv_byte, recv_num));
	debug(fflush(stdout));
    }
}

/*
 *  TDL_update_modify
 *
 *  Update only home node.  Each modified data is sent to the home
 *  node.
 */

void
TDL_update_modify(TDL_desc tdl_desc)
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {

	    assert(TDL_SEND_MLIST(tdl_desc)[i].num != NULL);

	    if (TDL_IS_HOME_PAGE(tdl_desc, i)) {
		int *sharing_rank = TDL_sharing_rank(tdl_desc, i);
		int num_shared = TDL_sharing_num(tdl_desc, i);
		int j, jj = 0;
		for (j = 0; j < num_shared; ++j) {
		    if (sharing_rank[j] != TDL_gThreadNum) {
			int k;
			for (k = 0; k < TDL_SEND_MLIST(tdl_desc)[i].num[j]; ++k) {
			    MPI_Status st;

			    debug(printf("[%d] MPI_Recv(%p (%p + %lx), %lu, MPI_CHAR, %d, TDL_UPDATE_MODIFY)\n",
					 TDL_gThreadNum, TDL_ADDR(TDL_SEND_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
					 TDL_BASE(tdl_desc), TDL_SEND_MLIST(tdl_desc)[i].offset[jj],
					 TDL_SEND_MLIST(tdl_desc)[i].size[jj], sharing_rank[j]));
			    debug(fflush(stdout));

			    MPI_Recv(TDL_ADDR(TDL_SEND_MLIST(tdl_desc)[i].offset[jj], tdl_desc),
				     TDL_SEND_MLIST(tdl_desc)[i].size[jj], MPI_CHAR,
				     sharing_rank[j], TDL_UPDATE_MODIFY,
				     TDL_gComm, &st);

			    jj++;
			}
		    }
		    else {
			jj += TDL_SEND_MLIST(tdl_desc)[i].num[j];
		    }
		}
		free(sharing_rank);
	    }
	    else {
		Send_el_list sl = TDL_SEND_LIST(tdl_desc)[i];
		while (sl) {
		    Send_el p = car_send_el(sl);

		    debug(printf("[%d] MPI_Send(%p (%p + %lx), %lu, MPI_CHAR, %ld, TDL_UPDATE_MODIFY)\n",
				 TDL_gThreadNum, TDL_ADDR(p->addr, tdl_desc),
				 TDL_BASE(tdl_desc), p->addr,
				 p->size, TDL_HOME(tdl_desc)[i]));
		    debug(fflush(stdout));

		    MPI_Send(TDL_ADDR(p->addr, tdl_desc),
			     p->size, MPI_CHAR,
			     TDL_HOME(tdl_desc)[i], TDL_UPDATE_MODIFY, TDL_gComm);

		    sl = cdr_send_el(sl);
		}
	    }
	}
/*	debug(printf("[%d] TDL_update_modify: i = %d: done\n", TDL_gThreadNum, i)); */
    }

#if 0 /* MPI2 or RHiNET */
    for (i = 0; i < TDL_NUM_PAGES(tdl_desc); ++i) {
	if (TDL_IS_SHARED(tdl_desc, i)) {
	    /*  If I'm not the home node. */
	    if (!(TDL_IS_HOME_PAGE(tdl_desc, i))) {
		Send_el_list sl = TDL_SEND_LIST(tdl_desc)[i];
		while (sl) {
		    Send_el p = car_send_el(sl);
#ifdef MPI2
#error Not implemented yet
		    MPI_Put(TDL_ADDR(p->addr, tdl_desc), p->size, MPI_CHAR,
			    TDL_HOME(tdl_desc)[i], p->addr, p->size, MPI_CHAR,
			    TDL_WIN(tdl_desc));
#elif defined(RHiNET)
#error Not implemented yet
		    /*  push_diff() */
#endif
		    sl = cdr_send_el(sl);
		}
	    }
	}
/*	debug(printf("[%d] TDL_update_modify: i = %d: done\n", TDL_gThreadNum, i)); */
    }
#endif
/*    debug(printf("[%d] TDL_update_modify: done\n", TDL_gThreadNum)); */
}

/*
 *  TDL_remap
 *
 *  TDL_remap() remaps a distributed array with a descriptor
 *  'tdl_old_desc' to that with 'tdl_new_desc', and changes the
 *  descriptor 'tdl_old_desc' by copying 'tdl_new_desc'.  Note
 *  that the base address (and the GID) of 'tdl_new_desc' is ignored.
 *
 *  ================================================================
 *  Now, the home node of each page sends a page to all newly shared
 *  nodes.  It should be rewritten such that every node having valid
 *  page including the home node sends the page to newly shared nodes
 *  respectively.
 *  ================================================================
 */

void
TDL_remap(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc)
{
    int i;

    /**  If the two dist. desc. are same, return immediately.  **/

    if (tdl_old_desc == tdl_new_desc) return;

    for (i = 0; i < TDL_NUM_PAGES(tdl_old_desc); ++i) {
	unsigned long start, end, offset;

	start = TDL_PAGE_OFFSET(i);
	offset = TDL_PAGE_OFFSET(i + 1);
	end = (offset < TDL_SIZE(tdl_old_desc)) ? offset : TDL_SIZE(tdl_old_desc);

	if (TDL_IS_HOME_PAGE(tdl_old_desc, i)) {
	    int j;
	    /*  multicast  */
	    for (j = 0; j < TDL_gNumThreads; ++j) {
		if (TDL_SHARING(tdl_old_desc)[i]  /**  Initialized  **/
		    && !TDL_IS_SHARED_WITH(tdl_old_desc, i, j)
		    && TDL_IS_SHARED_WITH(tdl_new_desc, i, j)) {

		    debug(printf("[%d] MPI_Send(%p (%p + %lx), %ld, MPI_CHAR, %d, TDL_REMAP)\n",
				 TDL_gThreadNum,
				 TDL_ADDR(start, tdl_old_desc),
				 TDL_BASE(tdl_old_desc), start,
				 end - start, j));
		    debug(fflush(stdout));

		    MPI_Send(TDL_ADDR(start, tdl_old_desc),
			     end - start, MPI_CHAR, j,
			     TDL_REMAP, TDL_gComm);
		}
	    }
	}
	else {
	    if (TDL_SHARING(tdl_old_desc)[i]  /**  Initialized  **/
		&& !TDL_IS_SHARED_WITH(tdl_old_desc, i, TDL_gThreadNum)
		&& TDL_IS_SHARED_WITH(tdl_new_desc, i, TDL_gThreadNum)) {
		MPI_Status status;

		debug(printf("[%d] MPI_Recv(%p (%p + %lx), %ld, MPI_CHAR, %ld, TDL_REMAP)\n",
			     TDL_gThreadNum,
			     TDL_ADDR(start, tdl_old_desc),
			     TDL_BASE(tdl_old_desc), start,
			     end - start, TDL_HOME(tdl_old_desc)[i]));
		debug(fflush(stdout));

		MPI_Recv(TDL_ADDR(start, tdl_old_desc),
			 end - start, MPI_CHAR,
			 TDL_HOME(tdl_old_desc)[i], TDL_REMAP,
			 TDL_gComm, &status);
	    }
	}
/*	debug(printf("[%d] TDL_remap: i = %d: done\n", TDL_gThreadNum, i)); */
    }

    /*  Copy descriptor. */

    _TDL_desc_copy_dist(tdl_new_desc, tdl_old_desc);

/*    debug(printf("[%d] TDL_remap: done\n", TDL_gThreadNum)); */
}

/*
 *  TDL_remap_home():
 *
 *  Remap home page only.  Return 1 if sharing is changed.
 *
 *  NOTE:
 *    TDL_remap_home() does not check TDL_REF() since it is not useful
 *    any more.  Be careful that TDL_remap_home() is not compatible
 *    with TDL_remap() because TDL_remap() still use TDL_REF() for
 *    duplication.
 */

static int
_TDL_remap_home(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc, int force)
{
    int i;
    int recv_pages = 0, send_pages = 0;
    int r = 0;

    /**  If the two dist. desc. are same, return immediately.  **/

    if (tdl_old_desc == tdl_new_desc) return r;

    for (i = 0; i < TDL_NUM_PAGES(tdl_old_desc); ++i) {

	/*  Transfer home page  */

	unsigned long start, end, offset;

	start = TDL_PAGE_OFFSET(i);
	offset = TDL_PAGE_OFFSET(i + 1);
	end = (offset < TDL_SIZE(tdl_old_desc)) ? offset : TDL_SIZE(tdl_old_desc);

	if (TDL_HOME(tdl_old_desc)[i] != TDL_HOME(tdl_new_desc)[i]
	    && (force || TDL_HOME(tdl_new_desc)[i] != TDL_PREV_HOME(tdl_old_desc)[i]))
	{
	    /*  page migration  */

	    if (TDL_IS_HOME_PAGE(tdl_old_desc, i)) {
		debug(printf("[%d] MPI_Send(%p (%p + %lx), %ld, MPI_CHAR, %ld, TDL_REMAP_HOME)\n",
			     TDL_gThreadNum,
			     TDL_ADDR(start, tdl_old_desc),
			     TDL_BASE(tdl_old_desc), start,
			     end - start, TDL_HOME(tdl_new_desc)[i]));
		debug(fflush(stdout));

		MPI_Send(TDL_ADDR(start, tdl_old_desc),
			 end - start, MPI_CHAR, TDL_HOME(tdl_new_desc)[i],
			 TDL_REMAP_HOME, TDL_gComm);

		debug(++send_pages);
	    }
	    if (TDL_IS_HOME_PAGE(tdl_new_desc, i)) {
		MPI_Status status;

		debug(printf("[%d] MPI_Recv(%p (%p + %lx), %ld, MPI_CHAR, %ld, TDL_REMAP_HOME)\n",
			     TDL_gThreadNum,
			     TDL_ADDR(start, tdl_old_desc),
			     TDL_BASE(tdl_old_desc), start,
			     end - start, TDL_HOME(tdl_old_desc)[i]));
		debug(fflush(stdout));

		MPI_Recv(TDL_ADDR(start, tdl_old_desc),
			 end - start, MPI_CHAR,
			 TDL_HOME(tdl_old_desc)[i], TDL_REMAP_HOME,
			 TDL_gComm, &status);

		debug(++recv_pages);
	    }

	    /*  In the case of initial page migration, unchange
		TDL_PREV_HOME to prevent unnecessary ping-pong
		detection. */
	    if (!force || TDL_PREV_HOME(tdl_old_desc)[i] != TDL_PAGE_UNDEF)
		TDL_PREV_HOME(tdl_old_desc)[i] = TDL_HOME(tdl_old_desc)[i];
	    TDL_HOME(tdl_old_desc)[i] = TDL_HOME(tdl_new_desc)[i];
	    TDL_SHARING(tdl_old_desc)[i] = TDL_SHARING(tdl_new_desc)[i];
	}
	else {
	    if (TDL_SHARING(tdl_old_desc)[i] != ((1 << TDL_HOME(tdl_old_desc)[i])
						 | TDL_SHARING(tdl_new_desc)[i])) {
		TDL_SHARING(tdl_old_desc)[i] = (1 << TDL_HOME(tdl_old_desc)[i])
		    | TDL_SHARING(tdl_new_desc)[i];
		r = 1;
	    }
	}
	/*  Beware of reference counter.  Copying may be necessary
	    if the reference counter is not 0.  */
    }

    if (send_pages > 0 || recv_pages > 0) {
	debug(printf("[%d] home page moved: base %p, send %d, recv %d, total num pages %d\n",
		     TDL_gThreadNum, TDL_BASE(tdl_old_desc), send_pages,
		     recv_pages, TDL_NUM_PAGES(tdl_old_desc)));
	debug(fflush(stdout));
    }

    /*  Copy descriptor. */
    /*  copy only reference information. */

    _TDL_desc_copy_sendrecv(tdl_new_desc, tdl_old_desc);

/*    debug(printf("[%d] TDL_remap_home: done\n", TDL_gThreadNum)); */

    return r;
}

int
TDL_remap_home(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc)
{
    return _TDL_remap_home(tdl_old_desc, tdl_new_desc, 0);
}

int
TDL_remap_home_force(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc)
{
    return _TDL_remap_home(tdl_old_desc, tdl_new_desc, 1);
}

/*
 *  TDL_remap_ref():
 *
 *  At first, remap home node.  Then, the home node sends only
 *  referenced data.
 */

void
TDL_remap_ref(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc)
{
    if (_TDL_remap_home(tdl_old_desc, tdl_new_desc, 0))
	TDL_desc_fix_ref(tdl_old_desc);
    TDL_update_ref(tdl_old_desc);
}

/*
 *  TDL_remap_modify():
 *
 *  At first, remap home node.  If necessary, fix descriptor.
 */

void
TDL_remap_modify(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc)
{
    if (_TDL_remap_home(tdl_old_desc, tdl_new_desc, 0))
	TDL_desc_fix_modify(tdl_old_desc);
}

/*
 *  TDL_remap_ref_modify():
 *
 *  At first, remap home node.  If necessary, fix descriptor.
 */

void
TDL_remap_ref_modify(TDL_desc tdl_old_desc, TDL_desc tdl_new_desc)
{
    if (_TDL_remap_home(tdl_old_desc, tdl_new_desc, 0)) {
	TDL_desc_fix_ref(tdl_old_desc);
	TDL_desc_fix_modify(tdl_old_desc);
    }
    TDL_update_ref(tdl_old_desc);
}

/*
 *  TDL_loop_schedule() assumes the following skeleton of 'for'
 *  loop:
 *
 *      for (i = start; i < end; i += step)
 */

void
TDL_loop_schedule(TDL_dist dist, int start, int end, int step,
		  int *l_start, int *l_end, int *l_step)
{
    int num_iters = (end - start + step - 1) / step;
    switch (dist) {
    case TDL_BLOCK:	/**  BLOCK distribution  **/
	*l_start = start + num_iters * TDL_gThreadNum / TDL_gNumThreads;
	*l_end = start + num_iters * (TDL_gThreadNum + 1) / TDL_gNumThreads;
	*l_step = step;
	break;
    case TDL_CYCLIC:	/**  CYCLIC distribution  **/
	*l_start = start + TDL_gThreadNum;
	*l_end = end;
	*l_step = step * TDL_gNumThreads;
	break;
    default:
	fprintf(stderr, "TDL_loop_index_to_local(): Not supported yet\n");
	exit(1);
    }    
    
    debug(printf("[%d] (start %d; end %d; step %d)\n", TDL_gThreadNum,
		 *l_start, *l_end, *l_step));
    debug(fflush(stdout));
}

int
TDL_barrier()
{
    return MPI_Barrier(TDL_gComm);
}

int
TDL_fence()
{
    return 0;

#if 0  /* MPI-2 */
    return MPI_Fence();
#endif
}
