/*
 *  $Id: tdl_allgather.c,v 1.1.1.1 2004-02-06 18:14:59 msato Exp $
 */

#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <mpi.h>

int
TDL_allgather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
	      void *recvbuf, int recvcount, MPI_Datatype recvtype,
	      int num_procs, int *group, int tag, MPI_Comm comm)
{
    int rank;
    MPI_Aint recv_ext;
    int recv_size, ssize, nbyte;
    MPI_Status st;
    char *sb = (char*)sendbuf, *rb = (char*)recvbuf;

    MPI_Comm_rank(comm, &rank);

    {
	int i;
	for (i = 0; i < num_procs; ++i)
	    if (group[i] == rank)
		break;
	rank = i;
	assert(rank != num_procs);
    }

    MPI_Type_extent(recvtype, &recv_ext);
    
    recv_size = recvcount * recv_ext;

    MPI_Type_size(sendtype, &ssize);
    nbyte = sendcount * ssize;
    memcpy(&rb[rank * recv_size], sb, nbyte);

    /**  Special case: #procs = 1  **/

    if (num_procs == 1) return 0;

    {
	int min = 0, max = num_procs;

	while (min < max - 1) {
	    int s = max - min;
	    int a = s / 2 + min;
	    int b = (s + 1) / 2 + min;

	    /**  exchange diagonal part  **/

	    if (rank < a) {
		int i;
		for (i = max - rank + min; i < max; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[i * recv_size], recvcount,
				 recvtype, group[i], tag,
				 comm, &st); 
		}
		for (i = b; i < max - rank + min; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[i * recv_size], recvcount,
				 recvtype, group[i], tag,
				 comm, &st);
		}
	    }
	    else if (rank >= b) {
		int i;
		for (i = max - rank + min; i < a; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[i * recv_size], recvcount,
				 recvtype, group[i], tag,
				 comm, &st);
		}
		for (i = min; i < max - rank + min; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[i * recv_size], recvcount,
				 recvtype, group[i], tag,
				 comm, &st);
		}
	    }

	    /**  exchange border part  **/

	    if (a != b) {
		if (rank == a) {
		    int i;
		    for (i = min; i < rank; ++i) {
			MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				     &rb[i * recv_size], recvcount, recvtype, group[i], tag,
				     comm, &st);
		    }
		    for (i = rank + 1; i < max; ++i) {
			MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				     &rb[i * recv_size], recvcount, recvtype, group[i], tag,
				     comm, &st);
		    }
		}
		else {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[a], tag,
				 &rb[a * recv_size], recvcount, recvtype, group[a], tag,
				 comm, &st);
		}
	    }

	    if (rank < a)
		max = a;
	    else if (rank >= b)
		min = b;
	    else
		max = min = 0;
	}
    }

    return 0;
}


int
TDL_allgatherv(void *sendbuf, int sendcount, MPI_Datatype sendtype,
	       void *recvbuf, int *recvcounts, int *displs, MPI_Datatype recvtype,
	       int num_procs, int *group, int tag, MPI_Comm comm)
{
    int rank;
    MPI_Aint recv_ext;
    int ssize, nbyte;
    MPI_Status st;
    char *sb = (char*)sendbuf, *rb = (char*)recvbuf;

    MPI_Comm_rank(comm, &rank);

    {
	int i;
	for (i = 0; i < num_procs; ++i)
	    if (group[i] == rank)
		break;
	rank = i;
	assert(rank != num_procs);
    }

    MPI_Type_extent(recvtype, &recv_ext);
    
    MPI_Type_size(sendtype, &ssize);
    nbyte = sendcount * ssize;
    memcpy(&rb[displs[rank] * recv_ext], sb, nbyte);

    /**  Special case: #procs = 1  **/

    if (num_procs == 1) return 0;

    {
	int min = 0, max = num_procs;

	while (min < max - 1) {
	    int s = max - min;
	    int a = s / 2 + min;
	    int b = (s + 1) / 2 + min;

	    /**  exchange diagonal part  **/

	    if (rank < a) {
		int i;
		for (i = max - rank + min; i < max; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[displs[i] * recv_ext], recvcounts[i], recvtype,
				 group[i], tag,
				 comm, &st); 
		}
		for (i = b; i < max - rank + min; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[displs[i] * recv_ext], recvcounts[i], recvtype,
				 group[i], tag,
				 comm, &st);
		}
	    }
	    else if (rank >= b) {
		int i;
		for (i = max - rank + min; i < a; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[displs[i] * recv_ext], recvcounts[i], recvtype,
				 group[i], tag,
				 comm, &st);
		}
		for (i = min; i < max - rank + min; ++i) {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				 &rb[displs[i] * recv_ext], recvcounts[i], recvtype,
				 group[i], tag,
				 comm, &st);
		}
	    }

	    /**  exchange border part  **/

	    if (a != b) {
		if (rank == a) {
		    int i;
		    for (i = min; i < rank; ++i) {
			MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				     &rb[displs[i] * recv_ext], recvcounts[i], recvtype,
				     group[i], tag,
				     comm, &st);
		    }
		    for (i = rank + 1; i < max; ++i) {
			MPI_Sendrecv(sb, sendcount, sendtype, group[i], tag,
				     &rb[displs[i] * recv_ext], recvcounts[i], recvtype,
				     group[i], tag,
				     comm, &st);
		    }
		}
		else {
		    MPI_Sendrecv(sb, sendcount, sendtype, group[a], tag,
				 &rb[displs[a] * recv_ext], recvcounts[a], recvtype,
				 group[a], tag,
				 comm, &st);
		}
	    }

	    if (rank < a)
		max = a;
	    else if (rank >= b)
		min = b;
	    else
		max = min = 0;
	}
    }

    return 0;
}


#if 0
int sendbuf[100];
int recvbuf[100];

main(argc, argv)
    int argc;
    char* argv[];
{
    int rank, size, *group;
    int i;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    group = (int*)malloc(size * sizeof(int));
    for (i = 0; i < size; ++i)
	group[i] = i;
    sendbuf[0] = rank;
    TDL_allgather(sendbuf, 1, MPI_INT,
		  recvbuf, 1, MPI_INT,
		  size, group, 100, MPI_COMM_WORLD);
    for (i = 0; i < size; ++i)
	if (recvbuf[i] != i)
	    fprintf(stderr, "[%d] ERROR: recvbuf[%d] = %d\n", rank, i, 
		    recvbuf[i]);

    MPI_Finalize();
}
#endif
