/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_mpi.c,v 1.8 2006-10-24 13:40:57 ojima Exp $";

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <netdb.h>
#include <fcntl.h>

#include "mpi.h"
#include "omsc.h"
#include "omsc_page.h"
#include "omsc_message.h"
#include "tlog-omsc.h"

/* tag */
#define OMSC_MSG_TAG 1
#define OMSC_RMA_CMD_TAG 2
#define OMSC_RMA_DATA_TAG_R 3
#define OMSC_RMA_DATA_TAG_W 5
#define OMSC_RAW_DATA_TAG 4

#define OMSC_MTU_SIZE 5000	/* 5K */

#define OMSC_WR  	0
#define OMSC_RD  	1

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

int omsc_debug_flag = 0;

int omsc_proc_id;	/* rank */
int omsc_num_proc;	/* number of processor */

static char omsc_my_hostname[MAXHOSTNAMELEN];

/* information for each node */
struct omsc_host {
    char hostname[MAXHOSTNAMELEN];
} omsc_hosts[SCASH_MAX_PE];

struct omsc_rma_header {
    short cmd;	 	/* read or write, write ack */
    short pe;		/* dst or src */
    int size;		/* message size */
    caddr_t addr;	/* address */
};

/* RMA server packet */
struct omsc_rma_pkt
{
    struct omsc_rma_pkt *next;	/* link */
    caddr_t addr;	/* address in this node */
    struct omsc_rma_header header;
};

static struct omsc_rma_pkt *omsc_rma_pkt_free = NULL;	/* free list */
static struct omsc_rma_pkt *omsc_rma_write = NULL; /* write request */
static struct omsc_rma_pkt *omsc_rma_read = NULL;  /* read request */
static struct omsc_rma_pkt *omsc_prefetch = NULL;  /* prefetch request */

struct omsc_msg_pkt {
    struct omsc_msg_pkt *next; /* link */
    int size;
    int dst;
    char buf[OMSC_MTU_SIZE];
};

pthread_t handler_tid;

static struct omsc_msg_pkt *omsc_msg_pkt_free = NULL;	/* free list */
static struct omsc_msg_pkt *omsc_msg_send_buf = NULL;	/* send buffer */
static struct omsc_msg_pkt *omsc_msg_recv_buf = NULL;	/* recv buffer */

static struct omsc_msg_pkt *omsc_msg_send = NULL; /* send request */
static struct omsc_msg_pkt *omsc_msg_recv = NULL; /* recv'ed packet */

pthread_mutex_t omsc_mutex; /* lock */
int omsc_mpi_handler_exit = FALSE;

#define LOCK  	pthread_mutex_lock(&omsc_mutex)
#define UNLOCK 	pthread_mutex_unlock(&omsc_mutex)

#define SCHED_YIELD sched_yield()

#define MAX_MAPFILE_NAME 256
char *omsc_tmp_prefix = "/tmp/";
static char omsc_mapfile_name[MAX_MAPFILE_NAME];

/* prototype */
static void omsc_fatal(char *msg,...);

static void *omsc_mpi_handler_main(void *dummy);
static struct omsc_msg_pkt *omsc_get_msg_pkt();
static void omsc_free_msg_pkt(struct omsc_msg_pkt *pkt);
static struct omsc_rma_pkt *omsc_get_rma_pkt();
static void omsc_free_rma_pkt(struct omsc_rma_pkt *pkt);

/*
 * initialize scash system
 */
void 
omsc_initialize(int argc, 
		char **argv)
{
    int provided;
    /* initialize hosting MPI */
    MPI_Init_thread(&argc,&argv,MPI_THREAD_FUNNELED,&provided);
    MPI_Comm_size(MPI_COMM_WORLD,&omsc_num_proc);
    MPI_Comm_rank(MPI_COMM_WORLD,&omsc_proc_id);

    scash_tlog_initialize();
    
    MPI_Barrier(MPI_COMM_WORLD);
}

void omsc_terminate(int status)
{
  int r;
  
  if(scash_debug) scash_prf("omsc_terminate ...");

  scash_barrier_sync(111);

  omsc_mpi_handler_exit = TRUE;

  if(scash.ps != 1){
      r = pthread_join(handler_tid,NULL);
      if(r != 0){
	  scash_prf("pthread_join r=%d\n",r);
      }
  }

  scash_tlog_finalize();

  if(scash_debug) scash_prf("omsc cleanup tmp files ...");
  unlink(omsc_mapfile_name);

  if(scash_debug) scash_prf("omsc MPI finalize ...");
  MPI_Finalize();

  if(scash.prefetch.access_history != NULL)
      free(scash.prefetch.access_history);

  if(scash_debug) scash_prf("now exit ...");
  exit(status);
}

void omsc_barrier()
{
    MPI_Barrier(MPI_COMM_WORLD);
}

void
omsc_init()
{
    char hostname[MAXHOSTNAMELEN];
    struct hostent *hp;
    int r,i;
    MPI_Status status;

    if(scash_debug) scash_prf("omsc_init ...");
    
    scash.pe = omsc_proc_id;
    scash.ps = omsc_num_proc;

    scash.proc_id = 0;	/* id in box */
    scash.node = omsc_proc_id; /* Id of box */
    scash.pid = getpid();

    scash.mtu = OMSC_MTU_SIZE;

    if (scash.ps > SCASH_MAX_PE) {
        scash_fatal("scash_score_init::# of PE is too large");
    }

    /* get hostname */
    r = gethostname(hostname,MAXHOSTNAMELEN);
    if(r < 0){
	perror("hostname");
	omsc_fatal("gethostname");
    }

    /* copy offical host name */
    hp = gethostbyname(hostname);
    if(hp == NULL){
	perror("gethostbyname");
	omsc_fatal("gethostbyname");
    }
    strcpy(omsc_my_hostname,hp->h_name);

    /* hostname */
    if(scash_debug) scash_prf("omsc_init: bcast hostname ...");
    if(omsc_proc_id != 0){	/* slave */
	MPI_Send(omsc_my_hostname,MAXHOSTNAMELEN,MPI_CHAR,0,0,MPI_COMM_WORLD);
    } else {	/* master */
	strcpy(omsc_hosts[0].hostname,omsc_my_hostname);
	for(i = 1; i < omsc_num_proc; i++)
	    MPI_Recv(omsc_hosts[i].hostname,MAXHOSTNAMELEN,MPI_CHAR,i,0,
		     MPI_COMM_WORLD,&status);
    }
    /* broadcast */
    MPI_Bcast(omsc_hosts,sizeof(struct omsc_host)*omsc_num_proc,
	      MPI_BYTE,0,MPI_COMM_WORLD);
    if(scash_debug) scash_prf("omsc_init: bcast hostname ... end");

    /* pe_node table, for non-SMP */
    for(i = 0; i < omsc_num_proc; i++)
	scash.pe_node[i] = i;

    if(scash_debug) scash_prf("omsc_init ... end");
    MPI_Barrier(MPI_COMM_WORLD);
}

/* non-SMP */
void omsc_init_mapfile(size_t size)
{
    int fd,i;

    sprintf(omsc_mapfile_name,"%s%s.%d",
	    omsc_tmp_prefix,SCASH_MAP_FILE,getpid());
    fd = open(omsc_mapfile_name,O_RDWR|O_CREAT|O_TRUNC,0777);
    if(fd < 0) 
	omsc_fatal("omsc_init_mapfile: can not create map file"); 

    /* initialize file */
    lseek(fd,size,0);
    if(write(fd,&i,1) != 1)
	omsc_fatal("omsc_init_mapfile:map file write");
    scash.map_fd = fd;
}

void omsc_start_mpi_handler()
{
    pthread_attr_t attr;
    int r, size;
    void* addr;

    if(scash.ps == 1) return;

    /* create lock */
    pthread_mutex_init(&omsc_mutex,NULL);

    /* start remote memory handler thread */
    pthread_attr_init(&attr);
    r = pthread_create(&handler_tid, &attr, omsc_mpi_handler_main, NULL);
    
    if(r != 0){
        omsc_fatal("cannot create handler thread");
    }
    if(scash_debug) scash_prf("omsc_mpi_handler start ...");
}

static struct omsc_msg_pkt *omsc_get_msg_pkt()
{
    struct omsc_msg_pkt *pkt;
#ifdef PROFILE
    u_int64 start, end;
    int diff;

    start = scash_rdtsc();
#endif    
    LOCK;
#ifdef PROFILE
    end = scash_rdtsc();
    diff = end - start;
    scash.profile.lock_get_time += diff;
    scash.profile.lock_get_cnt++;
#endif

    if((pkt = omsc_msg_pkt_free) == NULL){
	pkt = (struct omsc_msg_pkt *)malloc(sizeof(*pkt));
	if(pkt == NULL) omsc_fatal("cannot allocate omsc_msg_pkt");
    } else {
	omsc_msg_pkt_free = pkt->next;
    }
    UNLOCK;
    return pkt;
}

static void omsc_free_msg_pkt(struct omsc_msg_pkt *pkt)
{
/*
    LOCK;
    pkt->next = omsc_msg_pkt_free;
    omsc_msg_pkt_free = pkt;
    UNLOCK;
*/
    free(pkt);
}

static struct omsc_rma_pkt *omsc_get_rma_pkt()
{
    struct omsc_rma_pkt *pkt;
    
    LOCK;
    if((pkt = omsc_rma_pkt_free) == NULL){
	pkt = (struct omsc_rma_pkt *)malloc(sizeof(*pkt));
	if(pkt == NULL) omsc_fatal("cannot allocate omsc_rma_pkt");
    } else {
	omsc_rma_pkt_free = pkt->next;
    }
    UNLOCK;
    return pkt;
}

static void omsc_free_rma_pkt(struct omsc_rma_pkt *pkt)
{
/*
    pthread_mutex_lock(&omsc_mutex);
    pkt->next = omsc_rma_pkt_free;
    omsc_rma_pkt_free = pkt;
    pthread_mutex_unlock(&omsc_mutex);
*/
    free(pkt);
}

/* MPI handler process:
 *  note only this thread does call MPI functions 
 * because thread lib are not thread-safe.
 * - checking send request from DSM
 *   for msg send / write
 * - checking incomming packet 
 *    for recv/remote memory
 * - checking send done
 *    for send/write 
 */
static void *
omsc_mpi_handler_main(void *dummy)
{
    MPI_Status status;
    MPI_Request msg_recv_req, msg_send_req;
    MPI_Request rma_recv_req, rma_send_req;
    MPI_Request prefetch_req;
    int flag;
    struct omsc_rma_pkt *rma_pkt;  /* outstanding rma req */
    struct omsc_msg_pkt *msg_pkt;  /* outstanding msg req */
    struct omsc_msg_pkt *recv_pkt;  /* outstanding recv req */
    struct omsc_rma_pkt *prefetch_pkt; /* outstanding prefetch req */
    struct omsc_rma_header rma_header;
    int rma_recv_active;
    int page, rma_page;

    if(scash_debug) scash_prf("omsc_mpi_handler wakeup ...");
    msg_pkt = NULL;
    rma_pkt = NULL;
    recv_pkt = NULL;
    prefetch_pkt = NULL;
    rma_recv_active = FALSE;

again:
    /* Message Send */
    if(msg_pkt != NULL){
	/* if msg is outstanding, check request */
	MPI_Test(&msg_send_req,&flag,&status);
	if(flag){
	    //tlog_MPI_SEND_EVENT(); /* tlog */
	    LOCK;
	    omsc_msg_send = msg_pkt->next;  /* dequeue */
	    UNLOCK;
	    omsc_free_msg_pkt(msg_pkt);
	    msg_pkt = NULL;
	    /* scash_prf("mpi_handler: send msg done"); */
	    goto again;
	}
    } else {
	//if(omsc_msg_send != NULL){     /* checking send msg queue */
	if(omsc_msg_send != NULL && rma_pkt == NULL){
	    /* scash_prf("check omsc_msg_send=0x%x",omsc_msg_send); */
	    LOCK;
	    msg_pkt = omsc_msg_send;
	    UNLOCK;
	    if(msg_pkt != NULL){
#ifdef PROFILE
		if(scash.profile.get_msg_size){
		    scash.profile.total_msg_size += msg_pkt->size;
		    if(scash.profile.barrier_flag)
			scash.profile.barrier_msg_size += msg_pkt->size;
		    if(scash.profile.flush_flag){
			scash.profile.barrier_flush_size += msg_pkt->size;
			scash.profile.barrier_flush_cnt++;
		    }
		    if(scash.profile.inval_flag){
			scash.profile.barrier_inval_size += msg_pkt->size;
			scash.profile.barrier_inval_cnt++;
		    }
		    if(scash.profile.first_flag)
			scash.profile.barrier_1st_size += msg_pkt->size;
		}
#endif

		MPI_Isend(msg_pkt->buf,msg_pkt->size,
			  MPI_BYTE,msg_pkt->dst,
			  OMSC_MSG_TAG,MPI_COMM_WORLD,&msg_send_req);
	    }
	}
    }

    /* Remote Memory Access */
    if(rma_pkt != NULL){
	MPI_Test(&rma_send_req,&flag,&status);
	if(flag){
	    rma_page = (rma_pkt->addr - scash.shadow_addr_handle[scash.pe])/SCASH_PAGE_SIZE;

	    if(scash_tlog_flag) tlog_RMA_OUT(scash.pe);
	    LOCK;
	    if(rma_pkt == omsc_rma_write) 
		omsc_rma_write = rma_pkt->next;
	    else if(rma_pkt == omsc_rma_read)
		omsc_rma_read = rma_pkt->next;
	    else omsc_fatal("unknown RMA pkt");
	    UNLOCK;
	    /* request done */
	    omsc_free_rma_pkt(rma_pkt);
	    rma_pkt = NULL;

	    goto again;
	}
    } else {
	/* Remote Memory Write */
	if(omsc_rma_write != NULL){	/* check write queue */
	    LOCK;
	    rma_pkt = omsc_rma_write;
	    UNLOCK;
	    if(rma_pkt != NULL){
#ifdef PROFILE
		if(scash.profile.get_msg_size){
		    scash.profile.total_msg_size += 
			sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.barrier_flag)
			scash.profile.barrier_msg_size +=
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.flush_flag)
			scash.profile.barrier_flush_size += 
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.inval_flag)
			scash.profile.barrier_inval_size += 
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.first_flag)
			scash.profile.barrier_1st_size += 
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		}
#endif

		MPI_Isend(&rma_pkt->header,sizeof(struct omsc_rma_header),
			  MPI_BYTE,rma_pkt->header.pe,OMSC_RMA_CMD_TAG,
			  MPI_COMM_WORLD,&rma_send_req);
		MPI_Request_free(&rma_send_req);
		MPI_Isend(rma_pkt->addr,rma_pkt->header.size,
			 MPI_BYTE,rma_pkt->header.pe,OMSC_RMA_DATA_TAG_W,
			 MPI_COMM_WORLD,&rma_send_req);
	    }
	/* Remote Memory Read */
	} else if(omsc_rma_read != NULL){
	    LOCK;
	    rma_pkt = omsc_rma_read;
	    UNLOCK;
	    if(rma_pkt != NULL){
		rma_page = (rma_pkt->addr - scash.shadow_addr_handle[scash.pe])/SCASH_PAGE_SIZE;

#ifdef PROFILE
		if(scash.profile.get_msg_size){
		    scash.profile.total_msg_size += 
			sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.barrier_flag)
			scash.profile.barrier_msg_size +=
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.flush_flag)
			scash.profile.barrier_flush_size += 
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.inval_flag)
			scash.profile.barrier_inval_size += 
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		    if(scash.profile.first_flag)
			scash.profile.barrier_1st_size += 
			    sizeof(struct omsc_rma_header) + rma_pkt->header.size;
		}
#endif

		MPI_Isend(&rma_pkt->header,sizeof(struct omsc_rma_header),
			  MPI_BYTE,rma_pkt->header.pe,OMSC_RMA_CMD_TAG,
			  MPI_COMM_WORLD,&rma_send_req);
		MPI_Request_free(&rma_send_req);
		MPI_Irecv(rma_pkt->addr,rma_pkt->header.size,
			  MPI_BYTE,rma_pkt->header.pe,OMSC_RMA_DATA_TAG_R,
			  MPI_COMM_WORLD,&rma_send_req);
	    }
	}
    }

    if(recv_pkt == NULL){  
	/* activate asynchrous recv */
	recv_pkt = omsc_get_msg_pkt();
	MPI_Irecv(recv_pkt->buf,OMSC_MTU_SIZE,MPI_BYTE,
		  MPI_ANY_SOURCE,OMSC_MSG_TAG,MPI_COMM_WORLD,&msg_recv_req);
    }

    /* checking incoming msg packet (polling) */
    MPI_Test(&msg_recv_req,&flag,&status);
    if(flag){	
	/* put recv buffer */
	//tlog_MPI_RECV_EVENT(); /* tlog */
	LOCK;
	recv_pkt->next = omsc_msg_recv;
	omsc_msg_recv = recv_pkt;
	UNLOCK;
	recv_pkt = NULL;
	goto again;
    } 

    /* prefetch */
    if(prefetch_pkt != NULL){
	MPI_Test(&prefetch_req,&flag,&status);

	if(flag){ /* prefetch is done */
	    /* tlog_MPI_SEND_EVENT(); */
	    LOCK;
	    omsc_prefetch = prefetch_pkt->next;
	    omsc_free_rma_pkt(prefetch_pkt);
	    prefetch_pkt = NULL;
	    UNLOCK;

	    /* prefetch flag */
	    if(scash_pages[page].is_prefetching){
		//scash_page_RO(page);
		scash_page_touch(page);
		scash_pages[page].is_prefetching = 0;
	    }
	    goto again;
	}
    } else {
	if(omsc_prefetch != NULL){	/* check prefetch queue */
	    LOCK;
	    prefetch_pkt = omsc_prefetch;
	    UNLOCK;
	    if(prefetch_pkt != NULL){
		page = (prefetch_pkt->addr - scash.shadow_addr_handle[scash.pe])/SCASH_PAGE_SIZE;

#ifdef PROFILE
		if(scash.profile.get_msg_size){
		    scash.profile.total_msg_size += 
			sizeof(struct omsc_rma_header);
		    if(scash.profile.barrier_flag)
			scash.profile.barrier_msg_size +=
			    sizeof(struct omsc_rma_header);
		    if(scash.profile.flush_flag)
			scash.profile.barrier_flush_size += 
			    sizeof(struct omsc_rma_header);
		    if(scash.profile.inval_flag)
			scash.profile.barrier_inval_size += 
			    sizeof(struct omsc_rma_header);
		    if(scash.profile.first_flag)
			scash.profile.barrier_1st_size += 
			    sizeof(struct omsc_rma_header);
		}
#endif

		MPI_Isend(&prefetch_pkt->header,sizeof(struct omsc_rma_header),
			  MPI_BYTE,prefetch_pkt->header.pe,OMSC_RMA_CMD_TAG,
			  MPI_COMM_WORLD,&prefetch_req);
		MPI_Request_free(&prefetch_req);
		MPI_Irecv(prefetch_pkt->addr,prefetch_pkt->header.size,
			  MPI_BYTE,prefetch_pkt->header.pe,OMSC_RMA_DATA_TAG_R,
			  MPI_COMM_WORLD,&prefetch_req);
	    }
	}
    }

    if(!rma_recv_active){
	MPI_Irecv(&rma_header,sizeof(rma_header),MPI_BYTE,
		  MPI_ANY_SOURCE,OMSC_RMA_CMD_TAG,MPI_COMM_WORLD,
		  &rma_recv_req);
	rma_recv_active = TRUE;
    }

    /* checking incoming rma request (polling) */
    MPI_Test(&rma_recv_req,&flag,&status);
    if(flag){
	//tlog_MPI_RMA_EVENT(); /* tlog */
	/* send these msg in synchrous mode */
	switch(rma_header.cmd){
	case OMSC_WR:  	/* write */
	    MPI_Recv(rma_header.addr,rma_header.size,
		     MPI_BYTE,status.MPI_SOURCE,
		     OMSC_RMA_DATA_TAG_W,MPI_COMM_WORLD,&status);
	    break;
	case OMSC_RD:	/* read */
	    MPI_Send(rma_header.addr,rma_header.size,
		     MPI_BYTE,status.MPI_SOURCE,
		     OMSC_RMA_DATA_TAG_R,MPI_COMM_WORLD);
	    break;
	default:
	    omsc_fatal("unknown RMA pkt");
	}

	rma_page = (rma_header.addr - scash.shadow_addr_handle[scash.pe])/SCASH_PAGE_SIZE;
	scash_page_touch(rma_page);
	scash_page_cs_set(rma_page, status.MPI_SOURCE, FALSE);
	rma_recv_active = FALSE;
	
	goto again;
    }
    
    /* otherwise */
    SCHED_YIELD;

    if(omsc_msg_send != NULL || omsc_rma_write != NULL ||
       omsc_rma_read != NULL) goto again;

    if(!omsc_mpi_handler_exit) goto again;

    MPI_Barrier(MPI_COMM_WORLD);

    /* clean up */
    if(msg_pkt != NULL) MPI_Cancel(&msg_send_req);
    if(rma_pkt != NULL) MPI_Cancel(&rma_send_req);
    if(recv_pkt != NULL) MPI_Cancel(&msg_recv_req);
    if(rma_recv_active) MPI_Cancel(&rma_recv_req);

    if(scash_debug) scash_prf("omsc_mpi_handler terminated ...");
    return NULL;
}

void 
omsc_write(int dst_node, addrHandle dst_addr,addrHandle src_addr,int size)
{
    struct omsc_rma_pkt *pkt,*p;

    pkt = omsc_get_rma_pkt();
    pkt->header.cmd = OMSC_WR;
    pkt->header.pe = dst_node;
    pkt->header.size = size;
    pkt->header.addr = dst_addr;
    pkt->addr = src_addr;
    pkt->next = NULL;

    //scash_prf("OMSC_WRITE to %d : size %d", dst_node, size);
    /* put at tail */
    LOCK;
    if(omsc_rma_write == NULL) 
	omsc_rma_write = pkt;
    else {
	for(p = omsc_rma_write; p->next != NULL; p = p->next) /* */;
	p->next = pkt;
    }
    UNLOCK;
}

int 
omsc_is_write_done()
{
    int flag = FALSE;

    if(omsc_rma_write == NULL){
	/* check really NULL */
	LOCK; 
	if(omsc_rma_write == NULL) flag = TRUE; 
	UNLOCK;
    } else {
	SCHED_YIELD;
    }
    return flag;
}

int 
omsc_read(int src_node, addrHandle src_addr, addrHandle dst_addr, int size)
{
    struct omsc_rma_pkt *pkt,*p;
#ifdef PROFILE
    u_int64 start, end;
    int diff;
#endif

    pkt = omsc_get_rma_pkt();
    pkt->header.cmd = OMSC_RD;
    pkt->header.pe = src_node;
    pkt->header.size = size;
    pkt->header.addr = src_addr;
    pkt->addr = dst_addr;
    pkt->next = NULL;

    /* put at tail */
#ifdef PROFILE
    start = scash_rdtsc();
#endif
    LOCK;
#ifdef PROFILE
    end = scash_rdtsc();
    diff = end - start;
    scash.profile.lock_get_time += diff;
    scash.profile.lock_get_cnt++;
#endif
    if(omsc_rma_read == NULL) 
	omsc_rma_read = pkt;
    else {
	for(p = omsc_rma_read; p->next != NULL; p = p->next) /* */;
	p->next = pkt;
    }
    UNLOCK;
    return 0;
}

int 
omsc_is_read_done()
{
    int flag = FALSE;
#ifdef PROFILE
    u_int64 start, end;
    int diff;
#endif

    if(omsc_rma_read == NULL){
	/* check really NULL */
#ifdef PROFILE
	start = scash_rdtsc();
#endif
	LOCK;
#ifdef PROFILE
	end = scash_rdtsc();
	diff = end - start;
	scash.profile.lock_get_time += diff;
	scash.profile.lock_get_cnt++;
#endif
	if(omsc_rma_read == NULL) flag = TRUE; 
	UNLOCK;
    } else {
	SCHED_YIELD;
    }
    return flag;
}


int 
omsc_prefetch_page(int src_node, addrHandle src_addr, addrHandle dst_addr, int size)
{
    struct omsc_rma_pkt *pkt,*p;
#ifdef PROFILE
    u_int64 start, end;
    int diff;
#endif

    pkt = omsc_get_rma_pkt();
    pkt->header.cmd = OMSC_RD;
    pkt->header.pe = src_node;
    pkt->header.size = size;
    pkt->header.addr = src_addr;
    pkt->addr = dst_addr;
    pkt->next = NULL;

    /* put at tail */
#ifdef PROFILE
    start = scash_rdtsc();
#endif
    LOCK;
#ifdef PROFILE
    end = scash_rdtsc();
    diff = end - start;
    scash.profile.lock_get_time += diff;
    scash.profile.lock_get_cnt++;
#endif
    if(omsc_prefetch == NULL) 
	omsc_prefetch = pkt;
    else {
	for(p = omsc_prefetch; p->next != NULL; p = p->next) /* */;
	p->next = pkt;
    }
    UNLOCK;
    return 0;
}



int 
omsc_get_send_buf(int dst, caddr_t *bufpp, int size)
{
    struct omsc_msg_pkt *pkt;

    if(size > OMSC_MTU_SIZE) 
	omsc_fatal("pkt size %d > MTU",size);

    pkt = omsc_get_msg_pkt();
    pkt->size = size;
    pkt->dst = dst;
    pkt->next = NULL;
    *bufpp = pkt->buf;

    /* pkt->next = omsc_msg_send_buf; */
    omsc_msg_send_buf = pkt;
    
    return 0;
}

int 
omsc_trunc_buf(int size)
{
    omsc_msg_send_buf->size = size;
    return 0;
}

int 
omsc_send()
{
    struct omsc_msg_pkt *p;
#ifdef PROFILE
    u_int64 start, end;
    int diff;
    
    if(scash.profile.barrier_flag) scash.profile.barrier_msg_cnt++;
    start = scash_rdtsc();
#endif
    
    /* put at tail */
    LOCK;
#ifdef PROFILE
    end = scash_rdtsc();
    diff = end - start;
    scash.profile.lock_get_time += diff;
    scash.profile.lock_get_cnt++;
#endif
    if(omsc_msg_send == NULL) 
	omsc_msg_send = omsc_msg_send_buf;
    else {
	for(p = omsc_msg_send; p->next != NULL; p = p->next) /* */;
	p->next = omsc_msg_send_buf;
    }
    UNLOCK;
    return 0;
}

int 
omsc_is_send_done()
{
    int flag = FALSE;
    if(omsc_msg_send == NULL){
	/* check really NULL */
	LOCK; 
	if(omsc_msg_send == NULL) flag = TRUE; 
	UNLOCK;
    } else {
	SCHED_YIELD;
    }
    return flag;
}

int 
omsc_recv(caddr_t *msg, u_int *size) /* size is not used */
{
    struct omsc_msg_pkt *p,*pkt;

    pkt = NULL;
    /* pick the last one */
    if(omsc_msg_recv != NULL){
	LOCK;
	if(omsc_msg_recv != NULL){
	    if(omsc_msg_recv->next == NULL){
		pkt = omsc_msg_recv;
		omsc_msg_recv = NULL;
	    } else{
		for(p = omsc_msg_recv; p->next->next != NULL; p = p->next)/**/;
		pkt = p->next;
		p->next = NULL;
	    }
	} 
	UNLOCK;
    } 
    if(pkt == NULL) return ENOBUFS; 
    omsc_msg_recv_buf = pkt;
    *msg = omsc_msg_recv_buf->buf;
    return 0;
}

void 
omsc_release_buffer()
{
    omsc_free_msg_pkt(omsc_msg_recv_buf);
}

/* notify idle & busy */
void 
scash_idle()
{
    SCHED_YIELD;    /* nothing */
}

void 
scash_busy()
{
    SCHED_YIELD;    /* nothing */
}

/* lock memory if required */
void scash_memory_lock_init()
{
    int i;
    MPI_Status status;
    addrHandle addr_h;

    if(scash_debug) scash_prf("scash_memory_lock_init ...");

    if(omsc_proc_id != 0){	/* slave */
        addr_h = (addrHandle)scash.shadow_addr;
	MPI_Send((char *)&addr_h,sizeof(addr_h),MPI_BYTE,0,0,MPI_COMM_WORLD);
    } else {		/* master */
	scash.shadow_addr_handle[0] = scash.shadow_addr;
	for(i = 1; i < omsc_num_proc; i++){
	    MPI_Recv((char *)&addr_h,sizeof(addr_h),MPI_BYTE,i,0,
		     MPI_COMM_WORLD,&status);
	    scash.shadow_addr_handle[i] = addr_h;
	}
    }
    /* broadcast */
    MPI_Bcast(&scash.shadow_addr_handle,sizeof(addrHandle)*omsc_num_proc,
	      MPI_BYTE,0,MPI_COMM_WORLD);
    if(scash_debug) scash_prf("scash_memory_lock_init ... end");
}

void omsc_send_raw_data(int dst, char *p,int size)
{
    MPI_Send(p,size,MPI_BYTE,dst,OMSC_RAW_DATA_TAG,MPI_COMM_WORLD);
}

void omsc_recv_raw_data(int src, char *p,int size)
{
    MPI_Status status;
    MPI_Recv(p,size,MPI_BYTE,src,OMSC_RAW_DATA_TAG,MPI_COMM_WORLD,&status);
}

static void omsc_fatal(char *msg,...)
{
    va_list ap;
    va_start(ap,msg);
    fprintf(stderr,"OMSC_FATAL(%s): ",omsc_my_hostname);
    vfprintf(stderr,msg,ap);
    fprintf(stderr,"\n");
    va_end(ap);
    exit(1);
}

/* EOF */
